# This is for gecko-dev, which is a developer-oriented repo with
# release-train and inbound branches.

import os
import socket
hostname = socket.gethostname()

CVS_MANIFEST = """[{
"size": 1301484692,
"digest": "89df462d8d20f54402caaaa4e3c10aa54902a1d7196cdf86b7790b76e62d302ade3102dc3f7da4145dd832e6938b0472370ce6a321e0b3bcf0ad050937bd0e9a",
"algorithm": "sha512",
"filename": "mozilla-cvs-history.tar.bz2"
}]
"""

config = {
    "log_name": "beagle",
    "log_max_rotate": 99,
    "repos": [{
        "repo": "https://hg.mozilla.org/users/hwine_mozilla.com/repo-sync-tools",
        "vcs": "hg",
    }],
    "job_name": "beagle",
    "conversion_dir": "beagle",
    "initial_repo": {
        "repo": "https://hg.mozilla.org/mozilla-central",
        "revision": "default",
        "repo_name": "mozilla-central",
        "targets": [{
            "target_dest": "beagle/.git",
            "vcs": "git",
            "test_push": True,
        }, {
            "target_dest": "gitmo-beagle",
        }, {
            "target_dest": "github-beagle",
        }],
        "vcs": "hg",
        "branch_config": {
            "branches": {
                "default": "master",
            },
        },
    },
    "backup_dir": "/mnt/netapp/github_sync/aki/%s" % hostname,
    "cvs_manifest": CVS_MANIFEST,
    "cvs_history_tarball": "/home/pmoore/mozilla-cvs-history.tar.bz2",
    "env": {
        "PATH": "%(PATH)s:/usr/libexec/git-core",
    },
    "conversion_repos": [{
        "repo": "https://hg.mozilla.org/releases/mozilla-b2g18",
        "revision": "default",
        "repo_name": "mozilla-b2g18",
        "targets": [{
            "target_dest": "beagle/.git",
            "vcs": "git",
            "test_push": True,
        }, {
            "target_dest": "gitmo-beagle",
        }, {
            "target_dest": "github-beagle",
            "tag_config": {
                "tag_regexes": [
                    "^B2G_",
                ],
            },
        }],
        "vcs": "hg",
        "branch_config": {
            "branches": {
                "default": "b2g18",
            },
        },
        "tag_config": {
            "tag_regexes": [
                "^B2G_",
            ],
        },
        "mapper": {
            "url": "https://api.pub.build.mozilla.org/mapper",
            "project": "gecko-dev"
        },
    }, {
        "repo": "https://hg.mozilla.org/releases/mozilla-b2g26_v1_2",
        "revision": "default",
        "repo_name": "mozilla-b2g26_v1_2",
        "targets": [{
            "target_dest": "beagle/.git",
            "vcs": "git",
            "test_push": True,
        }, {
            "target_dest": "gitmo-beagle",
        }, {
            "target_dest": "github-beagle",
            "tag_config": {
                "tag_regexes": [
                    "^B2G_",
                ],
            },
        }],
        "vcs": "hg",
        "branch_config": {
            "branches": {
                "default": "b2g26_v1_2",
            },
        },
        "tag_config": {
            "tag_regexes": [
                "^B2G_",
            ],
        },
        "mapper": {
            "url": "https://api.pub.build.mozilla.org/mapper",
            "project": "gecko-dev"
        },
    }, {
        "repo": "https://hg.mozilla.org/releases/mozilla-b2g26_v1_2f",
        "revision": "default",
        "repo_name": "mozilla-b2g26_v1_2f",
        "targets": [{
            "target_dest": "beagle/.git",
            "vcs": "git",
            "test_push": True,
        }, {
            "target_dest": "gitmo-beagle",
        }, {
            "target_dest": "github-beagle",
            "tag_config": {
                "tag_regexes": [
                    "^B2G_",
                ],
            },
        }],
        "vcs": "hg",
        "branch_config": {
            "branches": {
                "default": "b2g26_v1_2f",
            },
        },
        "tag_config": {
            "tag_regexes": [
                "^B2G_",
            ],
        },
        "mapper": {
            "url": "https://api.pub.build.mozilla.org/mapper",
            "project": "gecko-dev"
        },
    }, {
        "repo": "https://hg.mozilla.org/releases/mozilla-b2g28_v1_3",
        "revision": "default",
        "repo_name": "mozilla-b2g28_v1_3",
        "targets": [{
            "target_dest": "beagle/.git",
            "vcs": "git",
            "test_push": True,
        }, {
            "target_dest": "gitmo-beagle",
        }, {
            "target_dest": "github-beagle",
            "tag_config": {
                "tag_regexes": [
                    "^B2G_",
                ],
            },
        }],
        "vcs": "hg",
        "branch_config": {
            "branches": {
                "default": "b2g28_v1_3",
            },
        },
        "tag_config": {
            "tag_regexes": [
                "^B2G_",
            ],
        },
        "mapper": {
            "url": "https://api.pub.build.mozilla.org/mapper",
            "project": "gecko-dev"
        },
    }, {
        "repo": "https://hg.mozilla.org/releases/mozilla-b2g28_v1_3t",
        "revision": "default",
        "repo_name": "mozilla-b2g28_v1_3t",
        "targets": [{
            "target_dest": "beagle/.git",
            "vcs": "git",
            "test_push": True,
        }, {
            "target_dest": "gitmo-beagle",
        }, {
            "target_dest": "github-beagle",
            "tag_config": {
                "tag_regexes": [
                    "^B2G_",
                ],
            },
        }],
        "vcs": "hg",
        "branch_config": {
            "branches": {
                "default": "b2g28_v1_3t",
            },
        },
        "tag_config": {
            "tag_regexes": [
                "^B2G_",
            ],
        },
        "mapper": {
            "url": "https://api.pub.build.mozilla.org/mapper",
            "project": "gecko-dev"
        },
    }, {
        "repo": "https://hg.mozilla.org/releases/mozilla-b2g30_v1_4",
        "revision": "default",
        "repo_name": "mozilla-b2g30_v1_4",
        "targets": [{
            "target_dest": "beagle/.git",
            "vcs": "git",
            "test_push": True,
        }, {
            "target_dest": "gitmo-beagle",
        }, {
            "target_dest": "github-beagle",
            "tag_config": {
                "tag_regexes": [
                    "^B2G_",
                ],
            },
        }],
        "vcs": "hg",
        "branch_config": {
            "branches": {
                "default": "b2g30_v1_4",
            },
        },
        "tag_config": {
            "tag_regexes": [
                "^B2G_",
            ],
        },
        "mapper": {
            "url": "https://api.pub.build.mozilla.org/mapper",
            "project": "gecko-dev"
        },
    }, {
        "repo": "https://hg.mozilla.org/releases/mozilla-b2g34_v2_1",
        "revision": "default",
        "repo_name": "mozilla-b2g34_v2_1",
        "targets": [{
            "target_dest": "beagle/.git",
            "vcs": "git",
            "test_push": True,
        }, {
            "target_dest": "gitmo-beagle",
        }, {
            "target_dest": "github-beagle",
        }],
        "vcs": "hg",
        "branch_config": {
            "branches": {
                "default": "b2g34_v2_1",
            },
        },
        "tag_config": {
            "tag_regexes": [
                "^B2G_",
            ],
        },
        "mapper": {
            "url": "https://api.pub.build.mozilla.org/mapper",
            "project": "gecko-dev"
        },
    }, {
        "repo": "https://hg.mozilla.org/releases/mozilla-b2g37_v2_2",
        "revision": "default",
        "repo_name": "mozilla-b2g37_v2_2",
        "targets": [{
            "target_dest": "beagle/.git",
            "vcs": "git",
            "test_push": True,
        }, {
            "target_dest": "gitmo-beagle",
        }, {
            "target_dest": "github-beagle",
        }],
        "vcs": "hg",
        "branch_config": {
            "branches": {
                "default": "b2g37_v2_2",
            },
        },
        "tag_config": {
            "tag_regexes": [
                "^B2G_",
            ],
        },
        "mapper": {
            "url": "https://api.pub.build.mozilla.org/mapper",
            "project": "gecko-dev"
        },
    }, {
        "repo": "https://hg.mozilla.org/releases/mozilla-b2g37_v2_2r",
        "revision": "default",
        "repo_name": "mozilla-b2g37_v2_2r",
        "targets": [{
            "target_dest": "beagle/.git",
            "vcs": "git",
            "test_push": True,
        }, {
            "target_dest": "gitmo-beagle",
        }, {
            "target_dest": "github-beagle",
        }],
        "vcs": "hg",
        "branch_config": {
            "branches": {
                "default": "b2g37_v2_2r",
            },
        },
        "tag_config": {
            "tag_regexes": [
                "^B2G_",
            ],
        },
        "mapper": {
            "url": "https://api.pub.build.mozilla.org/mapper",
            "project": "gecko-dev"
        },
    }, {
        "repo": "https://hg.mozilla.org/releases/mozilla-b2g32_v2_0",
        "revision": "default",
        "repo_name": "mozilla-b2g32_v2_0",
        "targets": [{
            "target_dest": "beagle/.git",
            "vcs": "git",
            "test_push": True,
        }, {
            "target_dest": "gitmo-beagle",
        }, {
            "target_dest": "github-beagle",
        }],
        "vcs": "hg",
        "branch_config": {
            "branches": {
                "default": "b2g32_v2_0",
            },
        },
        "tag_config": {
            "tag_regexes": [
                "^B2G_",
            ],
        },
        "mapper": {
            "url": "https://api.pub.build.mozilla.org/mapper",
            "project": "gecko-dev"
        },
    }, {
        "repo": "https://hg.mozilla.org/releases/mozilla-b2g32_v2_0m",
        "revision": "default",
        "repo_name": "mozilla-b2g32_v2_0m",
        "targets": [{
            "target_dest": "beagle/.git",
            "vcs": "git",
            "test_push": True,
        }, {
            "target_dest": "gitmo-beagle",
        }, {
            "target_dest": "github-beagle",
        }],
        "vcs": "hg",
        "branch_config": {
            "branches": {
                "default": "b2g32_v2_0m",
            },
        },
        "tag_config": {
            "tag_regexes": [
                "^B2G_",
            ],
        },
        "mapper": {
            "url": "https://api.pub.build.mozilla.org/mapper",
            "project": "gecko-dev"
        },
    }, {
        "repo": "https://hg.mozilla.org/releases/mozilla-b2g18_v1_1_0_hd",
        "revision": "default",
        "repo_name": "mozilla-b2g18_v1_1_0_hd",
        "targets": [{
            "target_dest": "beagle/.git",
            "vcs": "git",
            "test_push": True,
        }, {
            "target_dest": "gitmo-beagle",
        }, {
            "target_dest": "github-beagle",
        }],
        "vcs": "hg",
        "branch_config": {
            "branches": {
                "default": "b2g18_v1_1_0_hd",
            },
        },
        "tag_config": {
            "tag_regexes": [
                "^B2G_",
            ],
        },
        "mapper": {
            "url": "https://api.pub.build.mozilla.org/mapper",
            "project": "gecko-dev"
        },
    }, {
        "repo": "https://hg.mozilla.org/releases/mozilla-b2g18_v1_0_1",
        "revision": "default",
        "repo_name": "mozilla-b2g18_v1_0_1",
        "targets": [{
            "target_dest": "beagle/.git",
            "vcs": "git",
            "test_push": True,
        }, {
            "target_dest": "gitmo-beagle",
        }, {
            "target_dest": "github-beagle",
        }],
        "vcs": "hg",
        "branch_config": {
            "branches": {
                "default": "b2g18_v1_0_1",
            },
        },
        "tag_config": {
            "tag_regexes": [
                "^B2G_",
            ],
        },
        "mapper": {
            "url": "https://api.pub.build.mozilla.org/mapper",
            "project": "gecko-dev"
        },
    }, {
        "repo": "https://hg.mozilla.org/releases/mozilla-b2g18_v1_0_0",
        "revision": "default",
        "repo_name": "mozilla-b2g18_v1_0_0",
        "targets": [{
            "target_dest": "beagle/.git",
            "vcs": "git",
            "test_push": True,
        }, {
            "target_dest": "gitmo-beagle",
        }, {
            "target_dest": "github-beagle",
        }],
        "vcs": "hg",
        "branch_config": {
            "branches": {
                "default": "b2g18_v1_0_0",
            },
        },
        "tag_config": {
            "tag_regexes": [
                "^B2G_",
            ],
        },
        "mapper": {
            "url": "https://api.pub.build.mozilla.org/mapper",
            "project": "gecko-dev"
        },
    }, {
        "repo": "https://hg.mozilla.org/releases/mozilla-aurora",
        "revision": "default",
        "repo_name": "mozilla-aurora",
        "targets": [{
            "target_dest": "beagle/.git",
            "vcs": "git",
            "test_push": True,
        }, {
            "target_dest": "gitmo-beagle",
        }, {
            "target_dest": "github-beagle",
        }],
        "vcs": "hg",
        "branch_config": {
            "branches": {
                "default": "aurora",
            },
        },
        "tag_config": {
            "tag_regexes": [
                "^B2G_",
            ],
        },
        "mapper": {
            "url": "https://api.pub.build.mozilla.org/mapper",
            "project": "gecko-dev"
        },
    }, {
        "repo": "https://hg.mozilla.org/releases/mozilla-beta",
        "revision": "default",
        "repo_name": "mozilla-beta",
        "targets": [{
            "target_dest": "beagle/.git",
            "vcs": "git",
            "test_push": True,
        }, {
            "target_dest": "gitmo-beagle",
        }, {
            "target_dest": "github-beagle",
        }],
        "vcs": "hg",
        "branch_config": {
            "branches": {
                "default": "beta",
            },
            "branch_regexes": [
                "^GECKO[0-9_]*RELBRANCH$",
                "^MOBILE[0-9_]*RELBRANCH$",
            ],
        },
        "tag_config": {
            "tag_regexes": [
                "^(B2G|RELEASE_BASE)_",
            ],
        },
        "mapper": {
            "url": "https://api.pub.build.mozilla.org/mapper",
            "project": "gecko-dev"
        },
    }, {
        "repo": "https://hg.mozilla.org/releases/mozilla-release",
        "revision": "default",
        "repo_name": "mozilla-release",
        "targets": [{
            "target_dest": "beagle/.git",
            "vcs": "git",
            "test_push": True,
        }, {
            "target_dest": "gitmo-beagle",
        }, {
            "target_dest": "github-beagle",
        }],
        "vcs": "hg",
        "branch_config": {
            "branches": {
                "default": "release",
            },
            "branch_regexes": [
                "^GECKO[0-9_]*RELBRANCH$",
                "^MOBILE[0-9_]*RELBRANCH$",
            ],
        },
        "tag_config": {
            "tag_regexes": [
                "^(B2G|RELEASE_BASE)_",
            ],
        },
        "mapper": {
            "url": "https://api.pub.build.mozilla.org/mapper",
            "project": "gecko-dev"
        },
    }, {
        "repo": "https://hg.mozilla.org/releases/mozilla-esr17",
        "revision": "default",
        "repo_name": "mozilla-esr17",
        "targets": [{
            "target_dest": "beagle/.git",
            "vcs": "git",
            "test_push": True,
        }, {
            "target_dest": "gitmo-beagle",
        }, {
            "target_dest": "github-beagle",
        }],
        "vcs": "hg",
        "branch_config": {
            "branches": {
                "default": "esr17",
            },
        },
        "tag_config": {
            "tag_regexes": [
                "^B2G_",
            ],
        },
        "mapper": {
            "url": "https://api.pub.build.mozilla.org/mapper",
            "project": "gecko-dev"
        },
    }, {
        "repo": "https://hg.mozilla.org/releases/mozilla-esr31",
        "revision": "default",
        "repo_name": "mozilla-esr31",
        "targets": [{
            "target_dest": "beagle/.git",
            "vcs": "git",
            "test_push": True,
        }, {
            "target_dest": "gitmo-beagle",
        }, {
            "target_dest": "github-beagle",
        }],
        "vcs": "hg",
        "branch_config": {
            "branches": {
                "default": "esr31",
            },
            "branch_regexes": [
                "^GECKO[0-9]+esr_[0-9]+_RELBRANCH$",
            ],
        },
        "tag_config": {
            "tag_regexes": [
                "^B2G_",
            ],
        },
        "mapper": {
            "url": "https://api.pub.build.mozilla.org/mapper",
            "project": "gecko-dev"
        },
    }, {
        "repo": "https://hg.mozilla.org/releases/mozilla-esr38",
        "revision": "default",
        "repo_name": "mozilla-esr38",
        "targets": [{
            "target_dest": "beagle/.git",
            "vcs": "git",
            "test_push": True,
        }, {
            "target_dest": "gitmo-beagle",
        }, {
            "target_dest": "github-beagle",
        }],
        "vcs": "hg",
        "branch_config": {
            "branches": {
                "default": "esr38",
            },
            "branch_regexes": [
                "^GECKO[0-9]+esr_[0-9]+_RELBRANCH$",
            ],
        },
        "tag_config": {
            "tag_regexes": [
                "^B2G_",
            ],
        },
        "mapper": {
            "url": "https://api.pub.build.mozilla.org/mapper",
            "project": "gecko-dev"
        },
    }, {
        "repo": "https://hg.mozilla.org/integration/mozilla-inbound",
        "revision": "default",
        "repo_name": "mozilla-inbound",
        "targets": [{
            "target_dest": "beagle/.git",
            "vcs": "git",
            "test_push": True,
        }, {
            "target_dest": "gitmo-beagle",
        }, {
            "target_dest": "github-beagle",
        }],
        "vcs": "hg",
        "branch_config": {
            "branches": {
                "default": "inbound",
            },
        },
        "tag_config": {},
        "mapper": {
            "url": "https://api.pub.build.mozilla.org/mapper",
            "project": "gecko-dev"
        },
    }, {
        "repo": "https://hg.mozilla.org/integration/b2g-inbound",
        "revision": "default",
        "repo_name": "b2g-inbound",
        "targets": [{
            "target_dest": "beagle/.git",
            "vcs": "git",
            "test_push": True,
        }, {
            "target_dest": "gitmo-beagle",
        }, {
            "target_dest": "github-beagle",
        }],
        "vcs": "hg",
        "branch_config": {
            "branches": {
                "default": "b2g-inbound",
            },
        },
        "tag_config": {},
        "mapper": {
            "url": "https://api.pub.build.mozilla.org/mapper",
            "project": "gecko-dev"
        },
    }, {
        "repo": "https://hg.mozilla.org/integration/fx-team",
        "revision": "default",
        "repo_name": "fx-team",
        "targets": [{
            "target_dest": "beagle/.git",
            "vcs": "git",
            "test_push": True,
        }, {
            "target_dest": "gitmo-beagle",
        }, {
            "target_dest": "github-beagle",
        }],
        "vcs": "hg",
        "branch_config": {
            "branches": {
                "default": "fx-team",
            },
        },
        "tag_config": {},
        "mapper": {
            "url": "https://api.pub.build.mozilla.org/mapper",
            "project": "gecko-dev"
        },
    }],
    "remote_targets": {
        "github-beagle": {
            "repo": "git@github.com:mozilla/gecko-dev.git",
            "ssh_key": "~/.ssh/releng-github-id_rsa",
            "vcs": "git",
        },
        "gitmo-beagle": {
            "repo": "gitolite3@git.mozilla.org:integration/gecko-dev.git",
            "ssh_key": "~/.ssh/vcs-sync_rsa",
            "vcs": "git",
        },
    },

    "exes": {
        # bug 828140 - shut https warnings up.
        # http://kiln.stackexchange.com/questions/2816/mercurial-certificate-warning-certificate-not-verified-web-cacerts
        "hg": [os.path.join(os.getcwd(), "build", "venv", "bin", "hg"), "--config", "web.cacerts=/etc/pki/tls/certs/ca-bundle.crt"],
        "tooltool.py": [
            os.path.join(os.getcwd(), "build", "venv", "bin", "python"),
            os.path.join(os.getcwd(), "mozharness", "external_tools", "tooltool.py"),
        ],
    },

    "virtualenv_modules": [
        "bottle==0.11.6",
        "dulwich==0.9.0",
        "ordereddict==1.1",
        "hg-git==0.4.0-moz2",
        "mapper==0.1",
        "mercurial==2.6.3",
        "mozfile==0.9",
        "mozinfo==0.5",
        "mozprocess==0.11",
        "requests==2.8.1",
    ],
    "find_links": [
        "http://pypi.pvt.build.mozilla.org/pub",
        "http://pypi.pub.build.mozilla.org/pub",
    ],
    "pip_index": False,

    "default_notify_from": "developer-services+%s@mozilla.org" % hostname,
    "notify_config": [{
        "to": "releng-ops-trial@mozilla.com",
        "failure_only": False,
        "skip_empty_messages": True,
    }],

    # Disallow sharing, since we want pristine .hg and .git directories.
    "vcs_share_base": None,
    "hg_share_base": None,
    "default_actions": [
        'list-repos',
        'create-virtualenv',
        'update-stage-mirror',
        'update-work-mirror',
        'publish-to-mapper',
        'push',
        'combine-mapfiles',
        'notify',
    ],
}



# This is a template config file for panda android tests on production.
import socket
import os

config = {
    # Values for the foopies
    "exes": {
        'python': '/tools/buildbot/bin/python',
        'virtualenv': ['/tools/buildbot/bin/python', '/tools/misc-python/virtualenv.py'],
    },
    "run_file_names": {
        "preflight_talos": "remotePerfConfigurator.py",
        "talos": "run_tests.py",
    },
    "retry_url":  "http://talos-bundles.pvt.build.mozilla.org/zips/retry.zip",
    "verify_path":  "/builds/sut_tools/verify.py",
    "install_app_path":  "/builds/sut_tools/installApp.py",
    "talos_from_code_url": "https://hg.mozilla.org/%s/raw-file/%s/testing/talos/talos_from_code.py",
    "talos_json_url": "https://hg.mozilla.org/%s/raw-file/%s/testing/talos/talos.json",
    "datazilla_urls": ["https://datazilla.mozilla.org/talos"],
    "datazilla_authfile": os.path.join(os.getcwd(), "oauth.txt"),
    #remotePerfConfigurator.py options
    "preflight_talos_options": [
        "-v", "-e", "%(app_name)s",
        "-t", "%(hostname)s",
        "--branchName=%(talos_branch)s",
        "--resultsServer=graphs.mozilla.org",
        "--resultsLink=/server/collect.cgi",
        "--noChrome",
        "--symbolsPath=../symbols",
        "--remoteDevice=%(device_ip)s",
        "--sampleConfig=remote.config",
        "--output=local.yml",
        "--webServer=talos-remote.pvt.build.mozilla.org",
        "--browserWait=60"
    ],
    #run_tests.py options
    "talos_options": [
        "--noisy",
        "local.yml"
    ],
    "all_talos_suites": {
        "remote-tsvgx":  ["--activeTests=tsvgx", "--noChrome", '--tppagecycles', '10'],
        "remote-tp4m_nochrome":  ["--activeTests=tp4m", "--noChrome", "--rss"],
    },
    "find_links": [
        "http://pypi.pvt.build.mozilla.org/pub",
        "http://pypi.pub.build.mozilla.org/pub",
    ],
    "buildbot_json_path": "buildprops.json",
    "mobile_imaging_format": "http://mobile-imaging",
    "mozpool_assignee": socket.gethostname(),
    "default_actions": [
        'clobber',
        'read-buildbot-config',
        'download-and-extract',
        'create-virtualenv',
        'request-device',
        'run-test',
        'close-request',
    ],
    "default_blob_upload_servers": [
         "https://blobupload.elasticbeanstalk.com",
    ],
    "blob_uploader_auth_file" : os.path.join(os.getcwd(), "oauth.txt"),
}



#!/usr/bin/env python
# ***** BEGIN LICENSE BLOCK *****
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# ***** BEGIN LICENSE BLOCK *****
"""firefox_media_tests_buildbot.py

Author: Maja Frydrychowicz
"""
import copy
import glob
import os
import sys

sys.path.insert(1, os.path.dirname(sys.path[0]))

from mozharness.base.log import ERROR, DEBUG, INFO
from mozharness.base.script import PreScriptAction, PostScriptAction
from mozharness.mozilla.blob_upload import (
    BlobUploadMixin,
    blobupload_config_options
)
from mozharness.mozilla.buildbot import (
    TBPL_SUCCESS, TBPL_WARNING, TBPL_FAILURE
)
from mozharness.mozilla.testing.firefox_media_tests import (
    FirefoxMediaTestsBase, BUSTED, TESTFAILED, UNKNOWN, EXCEPTION, SUCCESS
)

buildbot_media_test_options = [
    [["--suite"],
     {"action": "store",
      "dest": "test_suite",
      "default": "media-tests",
      "help": "suite name",
      }],
]


class FirefoxMediaTestsBuildbot(FirefoxMediaTestsBase, BlobUploadMixin):

    def __init__(self):
        config_options = copy.deepcopy(blobupload_config_options)
        super(FirefoxMediaTestsBuildbot, self).__init__(
            config_options=config_options + buildbot_media_test_options,
            all_actions=['clobber',
                         'read-buildbot-config',
                         'checkout',
                         'download-and-extract',
                         'create-virtualenv',
                         'install',
                         'run-media-tests',
                         ],
        )
        c = self.config
        self.installer_url = c.get('installer_url')
        self.installer_path = c.get('installer_path')
        self.binary_path = c.get('binary_path')
        self.test_packages_url = c.get('test_packages_url')

    @PreScriptAction('create-virtualenv')
    def _pre_create_virtualenv(self, action):
        dirs = self.query_abs_dirs()
        requirements = os.path.join(dirs['abs_test_install_dir'],
                                    'config',
                                    'marionette_requirements.txt')
        if os.access(requirements, os.F_OK):
            self.register_virtualenv_module(requirements=[requirements],
                                            two_pass=True)
        super(FirefoxMediaTestsBuildbot, self)._pre_create_virtualenv(action)

    def query_abs_dirs(self):
        if self.abs_dirs:
            return self.abs_dirs
        dirs = super(FirefoxMediaTestsBuildbot, self).query_abs_dirs()
        dirs['abs_blob_upload_dir'] = os.path.join(dirs['abs_work_dir'],
                                                   'blobber_upload_dir')
        dirs['abs_test_install_dir'] = os.path.join(dirs['abs_work_dir'],
                                                    'tests')
        self.abs_dirs = dirs
        return self.abs_dirs

    def _query_cmd(self):
        """ Determine how to call firefox-media-tests """
        cmd = super(FirefoxMediaTestsBuildbot, self)._query_cmd()
        # configure logging
        dirs = self.query_abs_dirs()
        blob_upload_dir = dirs.get('abs_blob_upload_dir')
        cmd += ['--gecko-log', os.path.join(blob_upload_dir,
                                            'gecko.log')]
        cmd += ['--log-html', os.path.join(blob_upload_dir,
                                           'media_tests.html')]
        cmd += ['--log-mach', os.path.join(blob_upload_dir,
                                           'media_tests_mach.log')]

        test_suite = self.config.get('test_suite')
        test_manifest = None if test_suite != 'media-youtube-tests' else \
            os.path.join(dirs['firefox_media_dir'],
                         'firefox_media_tests',
                         'playback', 'youtube', 'manifest.ini')
        config_fmt_args = {
            'test_manifest': test_manifest,
        }

        if test_suite not in self.config["suite_definitions"]:
            self.fatal("%s is not defined in the config!" % test_suite)
        for s in self.config["suite_definitions"][test_suite]["options"]:
            cmd.append(s % config_fmt_args)

        return cmd

    def run_media_tests(self):
        status = super(FirefoxMediaTestsBuildbot, self).run_media_tests()
        if status == SUCCESS:
            tbpl_status = TBPL_SUCCESS
        else:
            tbpl_status = TBPL_FAILURE
        if status == TESTFAILED:
            tbpl_status = TBPL_WARNING
        self.buildbot_status(tbpl_status)

    @PostScriptAction('run-media-tests')
    def _collect_uploads(self, action, success=None):
        """ Copy extra (log) files to blob upload dir. """
        dirs = self.query_abs_dirs()
        log_dir = dirs.get('abs_log_dir')
        blob_upload_dir = dirs.get('abs_blob_upload_dir')
        if not log_dir or not blob_upload_dir:
            return
        self.mkdir_p(blob_upload_dir)
        # Move firefox-media-test screenshots into log_dir
        screenshots_dir = os.path.join(dirs['base_work_dir'],
                                       'screenshots')
        log_screenshots_dir = os.path.join(log_dir, 'screenshots')
        if os.access(log_screenshots_dir, os.F_OK):
            self.rmtree(log_screenshots_dir)
        if os.access(screenshots_dir, os.F_OK):
            self.move(screenshots_dir, log_screenshots_dir)

        # logs to upload: broadest level (info), error, screenshots
        uploads = glob.glob(os.path.join(log_screenshots_dir, '*'))
        log_files = self.log_obj.log_files
        log_level = self.log_obj.log_level

        def append_path(filename, dir=log_dir):
            if filename:
                uploads.append(os.path.join(dir, filename))

        append_path(log_files.get(ERROR))
        # never upload debug logs
        if log_level == DEBUG:
            append_path(log_files.get(INFO))
        else:
            append_path(log_files.get(log_level))
        # in case of SimpleFileLogger
        append_path(log_files.get('default'))
        for f in uploads:
            if os.access(f, os.F_OK):
                dest = os.path.join(blob_upload_dir, os.path.basename(f))
                self.copyfile(f, dest)


if __name__ == '__main__':
    media_test = FirefoxMediaTestsBuildbot()
    media_test.run_and_exit()



#!/usr/bin/env python
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import copy
import os
import re
import sys

# load modules from parent dir
sys.path.insert(1, os.path.dirname(sys.path[0]))

from mozharness.base.errors import BaseErrorList
from mozharness.base.log import ERROR, WARNING
from mozharness.base.script import PreScriptAction
from mozharness.base.vcs.vcsbase import MercurialScript
from mozharness.mozilla.blob_upload import BlobUploadMixin, blobupload_config_options
from mozharness.mozilla.testing.testbase import TestingMixin, testing_config_options


class B2GDesktopTest(BlobUploadMixin, TestingMixin, MercurialScript):
    test_suites = ('mochitest',
                   'reftest',)
    config_options = [
        [["--type"],
        {"action": "store",
         "dest": "test_type",
         "default": "browser",
         "help": "The type of tests to run",
        }],
        [["--browser-arg"],
        {"action": "store",
         "dest": "browser_arg",
         "default": None,
         "help": "Optional command-line argument to pass to the browser",
        }],
        [["--test-manifest"],
        {"action": "store",
         "dest": "test_manifest",
         "default": None,
         "help": "Path to test manifest to run",
        }],
        [["--test-suite"],
        {"action": "store",
         "dest": "test_suite",
         "type": "choice",
         "choices": test_suites,
         "help": "Which test suite to run",
        }],
        [["--total-chunks"],
        {"action": "store",
         "dest": "total_chunks",
         "help": "Number of total chunks",
        }],
        [["--this-chunk"],
        {"action": "store",
         "dest": "this_chunk",
         "help": "Number of this chunk",
        }]] + copy.deepcopy(testing_config_options) \
            + copy.deepcopy(blobupload_config_options)

    error_list = [
        {'substr': 'FAILED (errors=', 'level': ERROR},
        {'substr': r'''Could not successfully complete transport of message to Gecko, socket closed''', 'level': ERROR},
        {'substr': r'''Could not communicate with Marionette server. Is the Gecko process still running''', 'level': ERROR},
        {'substr': r'''Connection to Marionette server is lost. Check gecko''', 'level': ERROR},
        {'substr': 'Timeout waiting for marionette on port', 'level': ERROR},
        {'regex': re.compile(r'''(Timeout|NoSuchAttribute|Javascript|NoSuchElement|XPathLookup|NoSuchWindow|StaleElement|ScriptTimeout|ElementNotVisible|NoSuchFrame|InvalidElementState|NoAlertPresent|InvalidCookieDomain|UnableToSetCookie|InvalidSelector|MoveTargetOutOfBounds)Exception'''), 'level': ERROR},
    ]

    def __init__(self, options=[], require_config_file=False):
        super(B2GDesktopTest, self).__init__(
            config_options=self.config_options + copy.deepcopy(options),
            all_actions=['clobber',
                         'read-buildbot-config',
                         'pull',
                         'download-and-extract',
                         'create-virtualenv',
                         'install',
                         'run-tests'],
            default_actions=['clobber',
                             'download-and-extract',
                             'create-virtualenv',
                             'install',
                             'run-tests'],
            require_config_file=require_config_file,
            config={
                'require_test_zip': True,
            }
        )

        # these are necessary since self.config is read only
        c = self.config
        self.installer_url = c.get('installer_url')
        self.installer_path = c.get('installer_path')
        self.test_url = c.get('test_url')
        self.test_packages_url = c.get('test_packages_url')
        self.test_manifest = c.get('test_manifest')

        suite = self.config['test_suite']
        if suite not in self.test_suites:
            self.fatal("Don't know how to run --test-suite '%s'!" % suite)

    # TODO detect required config items and fail if not set

    def query_abs_dirs(self):
        if self.abs_dirs:
            return self.abs_dirs
        abs_dirs = super(B2GDesktopTest, self).query_abs_dirs()
        dirs = {}
        dirs['abs_blob_upload_dir'] = os.path.join(
                abs_dirs['abs_work_dir'], 'blobber_upload_dir')
        dirs['abs_tests_dir'] = os.path.join(abs_dirs['abs_work_dir'], 'tests')
        for d in self.test_suites + ('config', 'certs'):
            dirs['abs_%s_dir' % d] = os.path.join(
                    dirs['abs_tests_dir'], d)

        for key in dirs.keys():
            if key not in abs_dirs:
                abs_dirs[key] = dirs[key]
        self.abs_dirs = abs_dirs
        return self.abs_dirs

    @PreScriptAction('create-virtualenv')
    def _pre_create_virtualenv(self, action):
        dirs = self.query_abs_dirs()
        requirements = os.path.join(dirs['abs_config_dir'],
                                    'marionette_requirements.txt')
        self.register_virtualenv_module(requirements=[requirements],
                                        two_pass=True)

    def _query_abs_base_cmd(self, suite):
        dirs = self.query_abs_dirs()
        cmd = [self.query_python_path('python')]
        cmd.append(self.config["run_file_names"][suite])

        raw_log_file = os.path.join(dirs['abs_blob_upload_dir'],
                                    '%s_raw.log' % suite)
        error_summary_file = os.path.join(dirs['abs_blob_upload_dir'],
                                          '%s_errorsummary.log' % suite)

        str_format_values = {
            'application': self.binary_path,
            'test_manifest': self.test_manifest,
            'symbols_path': self.symbols_path,
            'gaia_profile': self.gaia_profile,
            'utility_path': os.path.join(dirs['abs_tests_dir'], 'bin'),
            'total_chunks': self.config.get('total_chunks'),
            'this_chunk': self.config.get('this_chunk'),
            'cert_path': dirs['abs_certs_dir'],
            'browser_arg': self.config.get('browser_arg'),
            'raw_log_file': raw_log_file,
            'error_summary_file': error_summary_file,
        }

        if suite not in self.config["suite_definitions"]:
            self.fatal("'%s' not defined in the config!" % suite)

        try_options, try_tests = self.try_args(suite)

        cmd.extend(self.query_options(self.config["suite_definitions"][suite]["options"],
                                      try_options,
                                      str_format_values=str_format_values))

        cmd.extend(self.query_tests_args(self.config["suite_definitions"][suite].get("tests"),
                                         try_tests,
                                         str_format_values=str_format_values))

        return cmd

    def download_and_extract(self):
        suite_category = self.config['test_suite']
        super(B2GDesktopTest, self).download_and_extract(suite_categories=[suite_category])

    def preflight_run_tests(self):
        super(B2GDesktopTest, self).preflight_run_tests()
        suite = self.config['test_suite']
        # set default test manifest by suite if none specified
        if not self.test_manifest:
            if suite == 'reftest':
                self.test_manifest = os.path.join('tests', 'layout',
                                                  'reftests', 'reftest.list')

        # set the gaia_profile
        self.gaia_profile = os.path.join(os.path.dirname(self.binary_path),
                                         'gaia', 'profile')

    def run_tests(self):
        """
        Run the tests
        """
        dirs = self.query_abs_dirs()

        error_list = self.error_list
        error_list.extend(BaseErrorList)

        suite = self.config['test_suite']
        if suite not in self.test_suites:
            self.fatal("Don't know how to run --test-suite '%s'!" % suite)

        cmd = self._query_abs_base_cmd(suite)

        cwd = dirs['abs_%s_dir' % suite]

        # TODO we probably have to move some of the code in
        # scripts/desktop_unittest.py and scripts/marionette.py to
        # mozharness.mozilla.testing.unittest so we can share it.
        # In the short term, I'm ok with some duplication of code if it
        # expedites things; please file bugs to merge if that happens.

        suite_name = [x for x in self.test_suites if x in self.config['test_suite']][0]
        if self.config.get('this_chunk'):
            suite = '%s-%s' % (suite_name, self.config['this_chunk'])
        else:
            suite = suite_name

        # bug 773703
        success_codes = None
        if suite_name == 'xpcshell':
            success_codes = [0, 1]

        env = {}
        if self.query_minidump_stackwalk():
            env['MINIDUMP_STACKWALK'] = self.minidump_stackwalk_path
        env['MOZ_UPLOAD_DIR'] = dirs['abs_blob_upload_dir']
        if not os.path.isdir(env['MOZ_UPLOAD_DIR']):
            self.mkdir_p(env['MOZ_UPLOAD_DIR'])
        env = self.query_env(partial_env=env)

        parser = self.get_test_output_parser(suite_name,
                                             config=self.config,
                                             log_obj=self.log_obj,
                                             error_list=error_list)

        return_code = self.run_command(cmd, cwd=cwd, env=env,
                                       output_timeout=1000,
                                       output_parser=parser,
                                       success_codes=success_codes)

        tbpl_status, log_level = parser.evaluate_parser(return_code,
                                                        success_codes=success_codes)
        parser.append_tinderboxprint_line(suite_name)

        self.buildbot_status(tbpl_status, level=log_level)
        self.log("The %s suite: %s ran with return status: %s" %
                 (suite_name, suite, tbpl_status), level=log_level)

if __name__ == '__main__':
    desktopTest = B2GDesktopTest()
    desktopTest.run_and_exit()




config = {
    "log_name": "updates_beta",
    "repo": {
        "repo": "https://hg.mozilla.org/build/tools",
        "revision": "default",
        "dest": "tools",
        "vcs": "hg",
    },
    "push_dest": "ssh://hg.mozilla.org/build/tools",
    "shipped-locales-url": "https://hg.mozilla.org/releases/mozilla-beta/raw-file/{revision}/browser/locales/shipped-locales",
    "ignore_no_changes": True,
    "ssh_user": "ffxbld",
    "ssh_key": "~/.ssh/ffxbld_rsa",
    "archive_domain": "archive.mozilla.org",
    "archive_prefix": "https://archive.mozilla.org/pub",
    "previous_archive_prefix": "https://archive.mozilla.org/pub",
    "download_domain": "download.mozilla.org",
    "balrog_url": "https://aus5.mozilla.org",
    "balrog_username": "ffxbld",
    "update_channels": {
        "beta": {
            "version_regex": r"^(\d+\.\d+(b\d+)?)$",
            "requires_mirrors": True,
            "patcher_config": "mozBeta-branch-patcher2.cfg",
            "update_verify_channel": "beta-localtest",
            "mar_channel_ids": [
                "firefox-mozilla-beta", "firefox-mozilla-release",
            ],
            "channel_names": ["beta", "beta-localtest", "beta-cdntest"],
            "rules_to_update": ["firefox-beta-cdntest", "firefox-beta-localtest"],
            "publish_rules": ["firefox-beta"],
        },
    },
    "balrog_use_dummy_suffix": False,
}



import os
import re
import urlparse
from collections import namedtuple

from mozharness.base.script import ScriptMixin
from mozharness.base.log import LogMixin, OutputParser, WARNING
from mozharness.base.errors import HgErrorList, VCSException
from mozharness.base.transfer import TransferMixin

HgtoolErrorList = [{
    'substr': 'abort: HTTP Error 404: Not Found',
    'level': WARNING,
}] + HgErrorList


class HgtoolParser(OutputParser):
    """
    A class that extends OutputParser such that it can find the "Got revision"
    string from hgtool.py output
    """

    got_revision_exp = re.compile(r'Got revision (\w+)')
    got_revision = None

    def parse_single_line(self, line):
        m = self.got_revision_exp.match(line)
        if m:
            self.got_revision = m.group(1)
        super(HgtoolParser, self).parse_single_line(line)


class HgtoolVCS(ScriptMixin, LogMixin, TransferMixin):
    def __init__(self, log_obj=None, config=None, vcs_config=None,
                 script_obj=None):
        super(HgtoolVCS, self).__init__()

        self.log_obj = log_obj
        self.script_obj = script_obj
        if config:
            self.config = config
        else:
            self.config = {}
        # vcs_config = {
        #  hg_host: hg_host,
        #  repo: repository,
        #  branch: branch,
        #  revision: revision,
        #  ssh_username: ssh_username,
        #  ssh_key: ssh_key,
        #  clone_by_revision: clone_by_revision,
        #  clone_with_purge: clone_with_purge,
        # }
        self.vcs_config = vcs_config
        self.hgtool = self.query_exe('hgtool.py', return_type='list')

    def ensure_repo_and_revision(self):
        """Makes sure that `dest` is has `revision` or `branch` checked out
        from `repo`.

        Do what it takes to make that happen, including possibly clobbering
        dest.
        """
        c = self.vcs_config
        for conf_item in ('dest', 'repo'):
            assert self.vcs_config[conf_item]
        dest = os.path.abspath(c['dest'])
        repo = c['repo']
        revision = c.get('revision')
        branch = c.get('branch')
        share_base = c.get('vcs_share_base', os.environ.get("HG_SHARE_BASE_DIR", None))
        clone_by_rev = c.get('clone_by_revision')
        clone_with_purge = c.get('clone_with_purge')
        env = {'PATH': os.environ.get('PATH')}
        if c.get('env'):
            env.update(c['env'])
        if share_base is not None:
            env['HG_SHARE_BASE_DIR'] = share_base
        if self._is_windows():
            # SYSTEMROOT is needed for 'import random'
            if 'SYSTEMROOT' not in env:
                env['SYSTEMROOT'] = os.environ.get('SYSTEMROOT', '')
            # HOME is needed for the 'hg help share' check
            if 'HOME' not in env:
                env['HOME'] = os.environ.get('HOME', '')

        cmd = self.hgtool[:]

        if clone_by_rev:
            cmd.append('--clone-by-revision')
        if branch:
            cmd.extend(['-b', branch])
        if revision:
            cmd.extend(['-r', revision])

        for base_mirror_url in self.config.get('hgtool_base_mirror_urls', self.config.get('vcs_base_mirror_urls', [])):
            bits = urlparse.urlparse(repo)
            mirror_url = urlparse.urljoin(base_mirror_url, bits.path)
            cmd.extend(['--mirror', mirror_url])

        for base_bundle_url in self.config.get('hgtool_base_bundle_urls', self.config.get('vcs_base_bundle_urls', [])):
            bundle_url = "%s/%s.hg" % (base_bundle_url, os.path.basename(repo))
            cmd.extend(['--bundle', bundle_url])

        cmd.extend([repo, dest])

        if clone_with_purge:
            cmd.append('--purge')

        parser = HgtoolParser(config=self.config, log_obj=self.log_obj,
                              error_list=HgtoolErrorList)
        retval = self.run_command(cmd, error_list=HgtoolErrorList, env=env, output_parser=parser)

        if retval != 0:
            raise VCSException("Unable to checkout")

        return parser.got_revision

    def query_pushinfo(self, repository, revision):
        """Query the pushdate and pushid of a repository/revision.
        This is intended to be used on hg.mozilla.org/mozilla-central and
        similar. It may or may not work for other hg repositories.
        """
        PushInfo = namedtuple('PushInfo', ['pushid', 'pushdate'])

        try:
            url = '%s/json-pushes?changeset=%s' % (repository, revision)
            self.info('Pushdate URL is: %s' % url)
            contents = self.retry(self.load_json_from_url, args=(url,))

            # The contents should be something like:
            # {
            #   "28537": {
            #    "changesets": [
            #     "1d0a914ae676cc5ed203cdc05c16d8e0c22af7e5",
            #    ],
            #    "date": 1428072488,
            #    "user": "user@mozilla.com"
            #   }
            # }
            #
            # So we grab the first element ("28537" in this case) and then pull
            # out the 'date' field.
            pushid = contents.iterkeys().next()
            self.info('Pushid is: %s' % pushid)
            pushdate = contents[pushid]['date']
            self.info('Pushdate is: %s' % pushdate)
            return PushInfo(pushid, pushdate)

        except Exception:
            self.exception("Failed to get push info from hg.mozilla.org")
            raise



#!/usr/bin/env python
# ***** BEGIN LICENSE BLOCK *****
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
# ***** END LICENSE BLOCK *****
"""sourcetool.py

Port of tools/buildfarm/utils/hgtool.py.

TODO: sourcetool.py currently ignores work_dir completely.
Maybe we should use it instead of dest ?
Maybe I need to rethink work_dir?
"""

import os
import sys

sys.path.insert(1, os.path.dirname(sys.path[0]))

from mozharness.base.config import parse_config_file
from mozharness.base.script import BaseScript
from mozharness.base.vcs.mercurial import MercurialVCS

# These variables allow us to create a sourcetool.py that will [hopefully]
# work with a selection of VCSs.
#
# To avoid needing to specify --vcs VCS, softlink sourcetool.py to
# VCStool.py (where VCS is the VCS you want to use, and is a member of
# VCS_CHOICES).
VCS_CHOICES = ['hg', ]
VCS_DEFAULT = None
VCS_REQUIRED_OPTION = "--vcs VCS "
VCS_NAME = os.path.basename(sys.argv[0]).replace('tool.py', '')
if VCS_NAME in VCS_CHOICES:
    VCS_DEFAULT = VCS_NAME
    VCS_REQUIRED_OPTION = ""
SOURCE_TOOL_USAGE = """Usage:
    %%prog [options] %srepo [dest]

    %%prog %s--repo REPOSITORY [options]

    %%prog [-h|--help]""" % (VCS_REQUIRED_OPTION, VCS_REQUIRED_OPTION)


# SourceTool {{{1
class SourceTool(BaseScript):
    # These options were chosen with an eye towards backwards
    # compatibility with the existing hgtool.
    #
    # TODO: get rid of env options, or at least remove HG from the names.
    config_options = [[
        ["--rev", "-r"],
        {"action": "store",
         "dest": "vcs_revision",
         "default": os.environ.get('HG_REV'),
         "help": "Specify which revision to update to."
         }
    ], [
        ["--branch", "-b"],
        {"action": "store",
         "dest": "vcs_branch",
         "default": os.environ.get('HG_BRANCH', 'default'),
         "help": "Specify which branch to update to."
         }
    ], [
        ["--vcs", ],
        {"action": "store",
         "type": "choice",
         "dest": "vcs",
         "default": VCS_DEFAULT,
         "choices": VCS_CHOICES,
         "help": "Specify which VCS to use."
         }
    ], [
        ["--props-file", "-p"],
        {"action": "store",
         "dest": "vcs_propsfile",
         "default": os.environ.get('PROPERTIES_FILE'),
         "help": "build json file containing revision information"
         }
    ], [
        # TODO --tbox and --no-tbox should DIAF once we fix bug 630538.
        ["--tbox", ],
        {"action": "store_true",
         "dest": "tbox_output",
         "default": bool(os.environ.get('PROPERTIES_FILE')),
         "help": "Output TinderboxPrint messages."
         }
    ], [
        ["--no-tbox", ],
        {"action": "store_false",
         "dest": "tbox_output",
         "help": "Don't output TinderboxPrint messages."
         }
    ], [
        ["--repo", ],
        {"action": "store",
         "dest": "vcs_repo",
         "help": "Specify the VCS repo."
         }
    ], [
        ["--dest", ],
        {"action": "store",
         "dest": "vcs_dest",
         "help": "Specify the destination directory (optional)"
         }
    ], [
        # TODO Are the shared options HG-specific?
        # I think there are, or we can create, similar behavior in other
        # VCSs.
        ["--shared-dir", '-s'],
        {"action": "store",
         "dest": "vcs_shared_dir",
         "default": os.environ.get('HG_SHARE_BASE_DIR'),
         "help": "clone to a shared directory"
         }
    ], [
        ["--allow-unshared-local-clones", ],
        {"action": "store_true",
         "dest": "vcs_allow_unshared_local_clones",
         "default": False,
         "help": "Allow unshared checkouts if --shared-dir is specified"
         }
    ], [
        ["--check-outgoing", ],
        {"action": "store_true",
         "dest": "vcs_strip_outgoing",
         "default": False,
         "help": "check for and clobber outgoing changesets"
         }
    ]]

    def __init__(self, require_config_file=False):
        BaseScript.__init__(self, config_options=self.config_options,
                            all_actions=['source', ],
                            usage=SOURCE_TOOL_USAGE,
                            require_config_file=require_config_file)

    def _pre_config_lock(self, rw_config):
        # This is a workaround for legacy compatibility with the original
        # hgtool.py.
        #
        # Since we need to read the buildbot json props, as well as parse
        # additional commandline arguments that aren't specified via
        # options, we call this function before locking the config.
        #
        # rw_config is the BaseConfig object that parsed the options;
        # self.config is the soon-to-be-locked runtime configuration.
        #
        # This is a powerful way to hack the config before locking;
        # we need to be careful not to abuse it.
        args = rw_config.args
        c = self.config
        if c.get('vcs') is None:
            self.fatal("Must specify --vcs!\n\n%s" %
                       rw_config.config_parser.format_help())
        if c.get('vcs_repo') is None:
            if len(args) not in (1, 2):
                self.fatal("""Invalid number of arguments!
You need to either specify --repo or specify it after the options:
%s""" % rw_config.config_parser.get_usage())

            self.config['vcs_repo'] = args[0]
        if len(args) == 2:
            self.config['vcs_dest'] = args[1]
        elif not self.config.get('vcs_dest'):
            self.config['vcs_dest'] = os.path.basename(self.config['vcs_repo'])

        # This is a buildbot-specific props file.
        if self.config.get('vcs_propsfile'):
            js = parse_config_file(self.config['vcs_propsfile'])
            if self.config.get('vcs_revision') is None:
                self.config['vcs_revision'] = js['sourcestamp']['revision']
            if self.config.get('vcs_branch') is None:
                self.config['vcs_branch'] = js['sourcestamp']['branch']

    def source(self):
        vcs_obj = None
        if self.config['vcs'] == 'hg':
            vcs_obj = MercurialVCS(
                log_obj=self.log_obj,
                #
                # Torn between creating a smaller, more flexible config per
                # helper object, or passing the read-only master config as
                # vcs_obj.config and creating a smaller vcs_obj.vcs_config.
                #
                # Deciding on the latter for now, while reserving the right
                # to change my mind later.
                config=self.config,
                vcs_config={
                    'repo': self.config['vcs_repo'],
                    'dest': self.config['vcs_dest'],
                    'branch': self.config.get('vcs_branch'),
                    'revision': self.config.get('vcs_revision'),
                    'vcs_share_base': self.config.get('vcs_shared_dir'),
                    'allow_unshared_local_clones': self.config.get('vcs_allow_unshared_local_clones'),
                    'halt_on_failure': self.config.get('halt_on_failure', True),
                }
            )
        else:
            self.fatal("I don't know how to handle vcs '%s'!" % self.config['vcs'])
        got_revision = vcs_obj.ensure_repo_and_revision()

        self.info("Got revision %s\n" % got_revision)

# __main__ {{{1
if __name__ == '__main__':
    source_tool = SourceTool()
    source_tool.run_and_exit()




# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""output formats for Talos"""

import filter
import json
import logging
import post_file
import time
import utils

# NOTE: we have a circular dependecy with output.py when we import results
import results as TalosResults

from StringIO import StringIO


def filesizeformat(bytes):
    """
    Format the value like a 'human-readable' file size (i.e. 13 KB, 4.1 MB, 102
    bytes, etc).
    """
    bytes = float(bytes)
    formats = ('B', 'KB', 'MB')
    for f in formats:
        if bytes < 1024:
            return "%.1f%s" % (bytes, f)
        bytes /= 1024
    return "%.1fGB" % bytes  # has to be GB


class Output(object):
    """abstract base class for Talos output"""

    @classmethod
    def check(cls, urls):
        """check to ensure that the urls are valid"""

    def __init__(self, results):
        """
        - results : TalosResults instance
        """
        self.results = results

    def __call__(self):
        """return list of results strings"""
        raise NotImplementedError("Abstract base class")

    def output(self, results, results_url, tbpl_output):
        """output to the results_url
        - results_url : http:// or file:// URL
        - results : list of results
        """

        # parse the results url
        results_url_split = utils.urlsplit(results_url)
        results_scheme, results_server, results_path, _, _ = results_url_split

        if results_scheme in ('http', 'https'):
            self.post(results, results_server, results_path, results_scheme,
                      tbpl_output)
        elif results_scheme == 'file':
            with open(results_path, 'w') as f:
                for result in results:
                    f.write("%s\n" % result)
        else:
            raise NotImplementedError(
                "%s: %s - only http://, https://, and file:// supported"
                % (self.__class__.__name__, results_url)
            )

    def post(self, results, server, path, scheme, tbpl_output):
        raise NotImplementedError("Abstract base class")

    @classmethod
    def shortName(cls, name):
        """short name for counters"""
        names = {"Working Set": "memset",
                 "% Processor Time": "%cpu",
                 "Private Bytes": "pbytes",
                 "RSS": "rss",
                 "XRes": "xres",
                 "Modified Page List Bytes": "modlistbytes",
                 "Main_RSS": "main_rss"}
        return names.get(name, name)

    @classmethod
    def isMemoryMetric(cls, resultName):
        """returns if the result is a memory metric"""
        memory_metric = ['memset', 'rss', 'pbytes', 'xres', 'modlistbytes',
                         'main_rss', 'content_rss']  # measured in bytes
        return bool([i for i in memory_metric if i in resultName])

    @classmethod
    def responsiveness_Metric(cls, val_list):
        return sum([float(x)*float(x) / 1000000.0 for x in val_list])

    @classmethod
    def v8_Metric(cls, val_list):
        results = [i for i, j in val_list]
        score = 100 * filter.geometric_mean(results)
        return score

    @classmethod
    def JS_Metric(cls, val_list):
        """v8 benchmark score"""
        results = [i for i, j in val_list]
        logging.info("javascript benchmark")
        return sum(results)

    @classmethod
    def CanvasMark_Metric(cls, val_list):
        """CanvasMark benchmark score (NOTE: this is identical to JS_Metric)"""
        results = [i for i, j in val_list]
        logging.info("CanvasMark benchmark")
        return sum(results)


class GraphserverOutput(Output):

    retries = 5  # number of times to attempt to contact graphserver
    info_format = ['title', 'testname', 'branch_name', 'sourcestamp',
                   'buildid', 'date']

    @classmethod
    def check(cls, urls):
        # ensure results_url link exists
        post_file.test_links(*urls)

    def __call__(self):
        """
        results to send to graphserver:
        construct all the strings of data, one string per test and one string
        per counter
        """

        result_strings = []

        info_dict = dict(
            title=self.results.title,
            date=self.results.date,
            branch_name=self.results.browser_config['branch_name'],
            sourcestamp=self.results.browser_config['sourcestamp'],
            buildid=self.results.browser_config['buildid'],
            browser_name=self.results.browser_config['browser_name'],
            browser_version=self.results.browser_config['browser_version']
        )

        for test in self.results.results:
            logging.debug("Working with test: %s", test.name())

            # get full name of test
            testname = test.name()
            if test.format == 'tpformat':
                # for some reason, we append the test extension to tp results
                # but not ts
                # http://hg.mozilla.org/build/talos/file/170c100911b6/talos
                # /run_tests.py#l176
                testname += test.extension()

            logging.debug("Generating results file: %s" % test.name())

            # HACK: when running xperf, we upload xperf counters to the graph
            # server but we do not want to
            # upload the test results as they will confuse the graph server
            if not (test.format == 'tpformat' and test.using_xperf):
                vals = []
                for result in test.results:
                    filtered_val = result.values(testname,
                                                 test.test_config['filters'])
                    vals.extend([[i['value'], j] for i, j in filtered_val])
                result_strings.append(self.construct_results(vals,
                                                             testname=testname,
                                                             **info_dict))

            # counter results
            for cd in test.all_counter_results:
                for counter_type, values in cd.items():
                    # get the counter name
                    counterName = '%s_%s' % (test.name(),
                                             self.shortName(counter_type))
                    if not values:
                        # failed to collect any data for this counter
                        logging.error(
                            "No results collected for: " + counterName
                        )
# NOTE: we are not going to enforce this warning for now as this happens too
# frequently: bugs 803413, 802475, 805925
#                        raise utils.TalosError("Unable to proceed with missing
# counter '%s'" % counterName)
# (jhammel: we probably should do this in e.g. results.py vs in
# graphserver-specific code anyway)

                    # exclude counters whose values are tuples (bad for
                    # graphserver)
                    if len(values) > 0 and isinstance(values[0], list):
                        print "Not uploading counter data for %s" % counterName
                        print values
                        continue

                    if test.mainthread() and 'mainthreadio' in counterName:
                        print ("Not uploading Mainthread IO data for %s"
                               % counterName)
                        print values
                        continue

                    # counter values
                    vals = [[x, 'NULL'] for x in values]

                    # append test name extension but only for tpformat tests
                    if test.format == 'tpformat':
                        counterName += test.extension()

                    info = info_dict.copy()
                    info['testname'] = counterName

                    # append the counter string
                    logging.info(
                        "Generating results file: %s" % counterName)
                    result_strings.append(self.construct_results(vals, **info))

        return result_strings

    def responsiveness_test(self, testname):
        """returns if the test is a responsiveness test"""
        # XXX currently this just looks for the string
        # 'responsiveness' in the test name.
        # It would be nice to be more declarative about this
        return 'responsiveness' in testname

    def construct_results(self, vals, testname, **info):
        """
        return results string appropriate to graphserver
        - vals: list of 2-tuples: [(val, page)
        - kwargs: info necessary for self.info_format interpolation
        see https://wiki.mozilla.org/Buildbot/Talos/DataFormat
        """

        info['testname'] = testname
        info_format = self.info_format
        responsiveness = self.responsiveness_test(testname)
        _type = 'VALUES'
        average = None
        if responsiveness:
            _type = 'AVERAGE'
            average = self.responsiveness_Metric([val for (val, page) in vals])
        elif testname.startswith('v8_7'):
            _type = 'AVERAGE'
            average = self.v8_Metric(vals)
        elif testname.startswith('kraken'):
            _type = 'AVERAGE'
            average = self.JS_Metric(vals)
        elif testname.startswith('tcanvasmark'):
            _type = 'AVERAGE'
            average = self.CanvasMark_Metric(vals)

        # ensure that we have all of the info data available
        missing = [i for i in info_format if i not in info]
        if missing:
            raise utils.TalosError("Missing keys: %s" % missing)
        info = ','.join([str(info[key]) for key in info_format])

        # write the data
        buffer = StringIO()
        buffer.write("START\n")
        buffer.write("%s\n" % _type)
        buffer.write('%s\n' % info)
        if average is not None:
            # write some kind of average
            buffer.write("%s\n" % average)
        else:
            for i, (val, page) in enumerate(vals):
                try:
                    buffer.write("%d,%.2f,%s\n" % (i, float(val), page))
                except ValueError:
                    logging.info(
                        "We expected a numeric value and recieved '%s' instead"
                        % val
                    )
                    pass

        buffer.write("END")
        return buffer.getvalue()

    def process_Request(self, post):
        """get links from the graphserver response"""
        links = ""
        for line in post.splitlines():
            if line.find("RETURN\t") > -1:
                line = line.replace("RETURN\t", "")
                links += line + '\n'
            logging.debug("process_Request line: %s", line)
        if not links:
            raise utils.TalosError("send failed, graph server says:\n%s"
                                   % post)
        return links

    def post(self, results, server, path, scheme, tbpl_output):
        """post results to the graphserver"""

        links = []
        wait_time = 5  # number of seconds between each attempt

        for index, data_string in enumerate(results):
            times = 0
            msg = ""
            while times < self.retries:
                logging.info(
                    "Posting result %d of %d to %s://%s%s, attempt %d",
                    index, len(results), scheme, server, path, times)
                try:
                    links.append(self.process_Request(
                        post_file.post_multipart(server, path,
                                                 files=[("filename",
                                                         "data_string",
                                                         data_string)])))
                    break
                except utils.TalosError, e:
                    msg = str(e)
                except Exception, e:
                    msg = str(e)
                times += 1
                time.sleep(wait_time)
                wait_time *= 2
            else:
                raise utils.TalosError(
                    "Graph server unreachable (%d attempts)\n%s"
                    % (self.retries, msg)
                )

        # add TBPL output
        self.add_tbpl_output(links, tbpl_output, server, scheme)

    def add_tbpl_output(self, links, tbpl_output, server, scheme):
        """
        add graphserver links such that TBPL can parse them.
        graphserver returns a response like:

          'tsvgr\tgraph.html#tests=[[224,113,14]]\ntsvgr\t2965.75\tgraph.html
          #tests=[[224,113,14]]\n'

        for each ts posted (tsvgr, in this case)
        """

        url_format = "%s://%s/%s"

        # XXX this will not work for multiple URLs :(
        tbpl_output.setdefault('graphserver', {})

        # XXX link_format to be deprecated; see
        # https://bugzilla.mozilla.org/show_bug.cgi?id=816634
        link_format = '<a href=\'%s\'>%s</a>'

        for response in links:

            # parse the response:
            # graphserver returns one of two responses.  For 'AVERAGE' payloads
            # graphserver returns a line
            # 'RETURN\t<test name>\t<value>\t<path segment>' :
            # http://hg.mozilla.org/graphs/file/8884ef9418bf/server/pyfomatic
            # /collect.py#l277
            # For 'VALUES' payloads, graphserver prepends an additional line
            # 'RETURN\t<test name>\t<path segment>' :
            # http://hg.mozilla.org/graphs/file/8884ef9418bf/server/pyfomatic
            # /collect.py#l274
            # see https://bugzilla.mozilla.org/show_bug.cgi?id=816634#c56 for
            # a more verbose explanation
            lines = [line.strip() for line in response.strip().splitlines()]
            assert len(lines) in (1, 2), """\
Should have one line for 'AVERAGE' payloads,
two lines for 'VALUES' payloads. You received:
%s""" % lines
            testname, result, path = lines[-1].split()
            if self.isMemoryMetric(testname):
                result = filesizeformat(result)

            # add it to the output
            url = url_format % (scheme, server, path)
            tbpl_output['graphserver'][testname] = {'url': url,
                                                    'result': result}

            # output to legacy TBPL; to be deprecated, see
            # https://bugzilla.mozilla.org/show_bug.cgi?id=816634
            linkName = '%s: %s' % (testname, result)
            print 'RETURN: %s' % link_format % (url, linkName)


class PerfherderOutput(Output):
    def __init__(self, results):
        Output.__init__(self, results)

    def output(self, results, results_url, tbpl_output):
        """output to the a file if results_url starts with file://
        - results : json instance
        - results_url : file:// URL
        """

        # parse the results url
        results_url_split = utils.urlsplit(results_url)
        results_scheme, results_server, results_path, _, _ = results_url_split

        # This is the output that treeherder expects to find when parsing the
        # log file
        logging.info("PERFHERDER_DATA: %s" % json.dumps(results))
        if results_scheme in ('file'):
            json.dump(results, file(results_path, 'w'), indent=2,
                      sort_keys=True)

    def post(self, results, server, path, scheme, tbpl_output):
        """conform to current code- not needed for perfherder"""
        pass

    def construct_results(self, vals, testname):
        if 'responsiveness' in testname:
            return self.responsiveness_Metric([val for (val, page) in vals])
        elif testname.startswith('v8_7'):
            return self.v8_Metric(vals)
        elif testname.startswith('kraken'):
            return self.JS_Metric(vals)
        elif testname.startswith('tcanvasmark'):
            return self.CanvasMark_Metric(vals)
        elif len(vals) > 1:
            return filter.geometric_mean([i for i, j in vals])
        else:
            return filter.mean([i for i, j in vals])

    def __call__(self):
        suites = []
        test_results = {
            'framework': {
                'name': 'talos',
            },
            'suites': suites,
        }

        for test in self.results.results:
            # serialize test results
            tsresult = None
            if not test.using_xperf:
                subtests = []
                suite = {
                    'name': test.name(),
                    'subtests': subtests,
                }
                suites.append(suite)
                vals = []
                replicates = {}

                # TODO: counters!!!! we don't have any, but they suffer the
                # same
                for result in test.results:
                    # XXX this will not work for manifests which list
                    # the same page name twice. It also ignores cycles
                    for page, val in result.raw_values():
                        if page == 'NULL':
                            page = test.name()
                            if tsresult is None:
                                tsresult = r = TalosResults.Results()
                                r.results = [{'index': 0, 'page': test.name(),
                                              'runs': val}]
                            else:
                                r = tsresult.results[0]
                                if r['page'] == test.name():
                                    r['runs'].extend(val)
                        replicates.setdefault(page, []).extend(val)

                tresults = [tsresult] if tsresult else test.results

                for result in tresults:
                    filtered_results = \
                        result.values(suite['name'],
                                      test.test_config['filters'])
                    vals.extend([[i['value'], j] for i, j in filtered_results])
                    for val, page in filtered_results:
                        if page == 'NULL':
                            # no real subtests
                            page = test.name()
                        subtest = {
                            'name': page,
                            'value': val['filtered'],
                            'replicates': replicates[page],
                        }
                        subtests.append(subtest)
                        if test.test_config.get('lower_is_better') is not None:
                            subtest['lowerIsBetter'] = \
                                test.test_config['lower_is_better']
                        if test.test_config.get('unit'):
                            subtest['unit'] = test.test_config['unit']

                # if there is more than one subtest, calculate a summary result
                if len(subtests) > 1:
                    suite['value'] = self.construct_results(
                        vals, testname=test.name())
                if test.test_config.get('lower_is_better') is not None:
                    suite['lowerIsBetter'] = \
                        test.test_config['lower_is_better']

            # counters results_aux data
            counter_subtests = []
            for cd in test.all_counter_results:
                for name, vals in cd.items():
                    # We want to add the xperf data as talos_counters
                    # exclude counters whose values are tuples (bad for
                    # graphserver)
                    if len(vals) > 0 and isinstance(vals[0], list):
                        continue

                    # mainthread IO is a list of filenames and accesses, we do
                    # not report this as a counter
                    if 'mainthreadio' in name:
                        continue

                    subtest = {
                        'name': name,
                        'value': 0.0,
                    }
                    counter_subtests.append(subtest)

                    if test.using_xperf:
                        subtest['value'] = vals[0]
                    else:
                        # calculate mean value
                        if len(vals) > 0:
                            varray = [float(v) for v in vals]
                            subtest['value'] = filter.mean(varray)
            if counter_subtests:
                suites.append({'name': test.name(),
                               'subtests': counter_subtests})
        return test_results

# available output formats
formats = {'datazilla_urls': PerfherderOutput,
           'results_urls': GraphserverOutput}



#!/usr/bin/env python
# ***** BEGIN LICENSE BLOCK *****
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# ***** BEGIN LICENSE BLOCK *****
"""firefox_media_tests.py

Author: Maja Frydrychowicz
"""
import copy
import os
import re

from mozharness.base.log import ERROR, WARNING
from mozharness.base.script import PreScriptAction
from mozharness.mozilla.testing.testbase import (TestingMixin,
                                                 testing_config_options)
from mozharness.mozilla.testing.unittest import TestSummaryOutputParserHelper
from mozharness.mozilla.vcstools import VCSToolsScript

BUSTED = 'busted'
TESTFAILED = 'testfailed'
UNKNOWN = 'unknown'
EXCEPTION = 'exception'
SUCCESS = 'success'

media_test_config_options = [
    [["--media-urls"],
     {"action": "store",
      "dest": "media_urls",
      "help": "Path to ini file that lists media urls for tests.",
      }],
    [["--profile"],
     {"action": "store",
      "dest": "profile",
      "default": None,
      "help": "Path to FF profile that should be used by Marionette",
      }],
    [["--test-timeout"],
     {"action": "store",
      "dest": "test_timeout",
      "default": 10000,
      "help": ("Number of seconds without output before"
                "firefox-media-tests is killed."
                "Set this based on expected time for all media to play."),
      }],
    [["--tests"],
     {"action": "store",
      "dest": "tests",
      "default": None,
      "help": ("Test(s) to run. Path to test_*.py or "
               "test manifest (*.ini)"),
      }],
    [["--e10s"],
     {"dest": "e10s",
      "action": "store_true",
      "default": False,
      "help": "Enable e10s when running marionette tests."
      }],
    [['--firefox-media-repo'], {
        'dest': 'firefox_media_repo',
        'default': 'https://github.com/mjzffr/firefox-media-tests.git',
        'help': 'which firefox_media_tests repo to use',
    }],
    [['--firefox-media-branch'], {
        'dest': 'firefox_media_branch',
        'default': 'master',
        'help': 'which branch to use for firefox_media_tests',
    }],
    [['--firefox-media-rev'], {
        'dest': 'firefox_media_rev',
        'help': 'which firefox_media_tests revision to use',
    }],
    [['--firefox-ui-repo'], {
        'dest': 'firefox_ui_repo',
        'default': 'https://github.com/mozilla/firefox-ui-tests.git',
        'help': 'which firefox_ui_tests repo to use',
    }],
    [['--firefox-ui-branch'], {
        'dest': 'firefox_ui_branch',
        'default': 'master',
        'help': 'which branch to use for firefox_ui_tests',
    }],
    [['--firefox-ui-rev'], {
        'dest': 'firefox_ui_rev',
        'help': 'which firefox_ui_tests revision to use',
    }],
] + (copy.deepcopy(testing_config_options))


class JobResultParser(TestSummaryOutputParserHelper):
    """ Parses test output to determine overall result."""
    def __init__(self, **kwargs):
        super(JobResultParser, self).__init__(**kwargs)
        self.return_code = 0
        # External-resource errors that should not count as test failures
        self.exception_re = re.compile(r'^TEST-UNEXPECTED-ERROR.*'
                                       r'TimeoutException: Error loading page,'
                                       r' timed out')
        self.exceptions = []

    def parse_single_line(self, line):
        super(JobResultParser, self).parse_single_line(line)
        if self.exception_re.match(line):
            self.exceptions.append(line)

    @property
    def status(self):
        status = UNKNOWN
        if self.passed and self.failed == 0:
            status = SUCCESS
        elif self.exceptions:
            status = EXCEPTION
        elif self.failed:
            status = TESTFAILED
        elif self.return_code:
            status = BUSTED
        return status


class FirefoxMediaTestsBase(TestingMixin, VCSToolsScript):
    job_result_parser = None

    error_list = [
        {'substr': 'FAILED (errors=', 'level': WARNING},
        {'substr': r'''Could not successfully complete transport of message to Gecko, socket closed''', 'level': ERROR},
        {'substr': r'''Connection to Marionette server is lost. Check gecko''', 'level': ERROR},
        {'substr': 'Timeout waiting for marionette on port', 'level': ERROR},
        {'regex': re.compile(r'''(TEST-UNEXPECTED|PROCESS-CRASH|CRASH|ERROR|FAIL)'''), 'level': ERROR},
        {'regex': re.compile(r'''(\b\w*Exception)'''), 'level': ERROR},
        {'regex': re.compile(r'''(\b\w*Error)'''), 'level': ERROR},
     ]

    def __init__(self, config_options=None, all_actions=None,
                 default_actions=None, **kwargs):
        self.config_options = media_test_config_options + (config_options or [])
        actions = [
            'clobber',
            'checkout',
            'create-virtualenv',
            'run-media-tests',
        ]
        super(FirefoxMediaTestsBase, self).__init__(
            config_options=self.config_options,
            all_actions=all_actions or actions,
            default_actions=default_actions or actions,
            **kwargs
        )
        c = self.config
        self.media_urls = c.get('media_urls')
        self.profile = c.get('profile')
        self.test_timeout = int(c.get('test_timeout'))
        self.tests = c.get('tests')
        self.e10s = c.get('e10s')

    @PreScriptAction('create-virtualenv')
    def _pre_create_virtualenv(self, action):
        dirs = self.query_abs_dirs()
        requirements_file = os.path.join(dirs['firefox_media_dir'],
                                         'requirements.txt')
        if os.path.isfile(requirements_file):
            self.register_virtualenv_module(requirements=[requirements_file])
        self.register_virtualenv_module(name='firefox-ui-tests',
                                        url=dirs['firefox_ui_dir'])
        self.register_virtualenv_module(name='firefox-media-tests',
                                        url=dirs['firefox_media_dir'])

    def query_abs_dirs(self):
        if self.abs_dirs:
            return self.abs_dirs
        abs_dirs = super(FirefoxMediaTestsBase, self).query_abs_dirs()
        dirs = {
            'firefox_media_dir': os.path.join(abs_dirs['abs_work_dir'],
                                              'firefox-media-tests')
        }
        dirs['firefox_ui_dir'] = os.path.join(dirs['firefox_media_dir'],
                                              'firefox-ui-tests')
        abs_dirs.update(dirs)
        self.abs_dirs = abs_dirs
        return self.abs_dirs

    @PreScriptAction('checkout')
    def _pre_checkout(self, action):
        super(FirefoxMediaTestsBase, self)._pre_checkout(action)
        c = self.config
        dirs = self.query_abs_dirs()
        self.firefox_media_vc = {
            'branch': c['firefox_media_branch'],
            'repo': c['firefox_media_repo'],
            'revision': c['firefox_media_rev'],
            'dest': dirs['firefox_media_dir'],
        }
        self.firefox_ui_vc = {
            'branch': c['firefox_ui_branch'],
            'repo': c['firefox_ui_repo'],
            'revision': c['firefox_ui_rev'],
            'dest': dirs['firefox_ui_dir']
        }

    def checkout(self):
        revision = self.vcs_checkout(vcs='gittool', **self.firefox_media_vc)
        if revision:
            self.vcs_checkout(vcs='gittool', **self.firefox_ui_vc)

    def _query_cmd(self):
        """ Determine how to call firefox-media-tests """
        if not self.binary_path:
            self.fatal("Binary path could not be determined. "
                       "Should be set by default during 'install' action.")
        dirs = self.query_abs_dirs()
        venv_python_path = self.query_python_path()
        runner_script = os.path.join(dirs['firefox_media_dir'],
                                     'media_test_harness',
                                     'runtests.py')
        cmd = [venv_python_path, runner_script]
        cmd += ['--binary', self.binary_path]
        if self.symbols_path:
            cmd += ['--symbols-path', self.symbols_path]
        if self.media_urls:
            cmd += ['--urls', self.media_urls]
        if self.profile:
            cmd += ['--profile', self.profile]
        if self.tests:
            cmd.append(self.tests)
        if self.e10s:
            cmd.append('--e10s')

        return cmd

    def run_media_tests(self):
        cmd = self._query_cmd()
        self.job_result_parser = JobResultParser(
            config=self.config,
            log_obj=self.log_obj,
            error_list=self.error_list
        )

        env = self.query_env()
        if (not os.environ.get('MINIDUMP_STACKWALK') and
                self.query_minidump_stackwalk()):
            env['MINIDUMP_STACKWALK'] = self.minidump_stackwalk_path

        return_code = self.run_command(
            cmd,
            output_timeout=self.test_timeout,
            output_parser=self.job_result_parser,
            env=env
        )
        self.job_result_parser.return_code = return_code
        return self.job_result_parser.status



# This is a template config file for b2g emulator unittest production.
import os

config = {
    # mozharness options
    "application": "b2g",
    "busybox_url": "https://api.pub.build.mozilla.org/tooltool/sha512/0748e900821820f1a42e2f1f3fa4d9002ef257c351b9e6b78e7de0ddd0202eace351f440372fbb1ae0b7e69e8361b036f6bd3362df99e67fc585082a311fc0df",
    "xre_url": "https://api.pub.build.mozilla.org/tooltool/sha512/dc9503b21c87b5a469118746f99e4f41d73888972ce735fa10a80f6d218086da0e3da525d9a4cd8e4ea497ec199fef720e4a525873d77a1af304ac505e076462",
    "tooltool_cache": "/builds/tooltool_cache",

    "exes": {
        'python': '/tools/buildbot/bin/python',
        'virtualenv': ['/tools/buildbot/bin/python', '/tools/misc-python/virtualenv.py'],
        'tooltool.py': "/tools/tooltool.py",
    },

    "find_links": [
        "http://pypi.pvt.build.mozilla.org/pub",
        "http://pypi.pub.build.mozilla.org/pub",
    ],
    "pip_index": False,

    "buildbot_json_path": "buildprops.json",

    "default_actions": [
        'clobber',
        'read-buildbot-config',
        'download-and-extract',
        'create-virtualenv',
        'install',
        'run-tests',
    ],
    "download_symbols": "ondemand",
    "download_minidump_stackwalk": True,
    "default_blob_upload_servers": [
        "https://blobupload.elasticbeanstalk.com",
    ],
    "blob_uploader_auth_file": os.path.join(os.getcwd(), "oauth.txt"),

    "run_file_names": {
        "jsreftest": "runreftestb2g.py",
        "mochitest": "runtestsb2g.py",
        "mochitest-chrome": "runtestsb2g.py",
        "reftest": "runreftestb2g.py",
        "crashtest": "runreftestb2g.py",
        "xpcshell": "runtestsb2g.py",
        "cppunittest": "remotecppunittests.py"
    },
    "suite_definitions": {
        "cppunittest": {
            "options": [
                "--dm_trans=adb",
                "--symbols-path=%(symbols_path)s",
                "--xre-path=%(xre_path)s",
                "--addEnv",
                "LD_LIBRARY_PATH=/vendor/lib:/system/lib:/system/b2g",
                "--with-b2g-emulator=%(b2gpath)s",
                 "--emulator=%(emulator)s",
                "."
            ],
            "run_filename": "remotecppunittests.py",
            "testsdir": "cppunittest"
        },
        "crashtest": {
            "options": [
                "--adbpath=%(adbpath)s",
                "--b2gpath=%(b2gpath)s",
                "--emulator=%(emulator)s",
                "--emulator-res=800x1000",
                "--logdir=%(logcat_dir)s",
                "--remote-webserver=%(remote_webserver)s",
                "--ignore-window-size",
                "--xre-path=%(xre_path)s",
                "--symbols-path=%(symbols_path)s",
                "--busybox=%(busybox)s",
                "--total-chunks=%(total_chunks)s",
                "--this-chunk=%(this_chunk)s",
                "--suite=crashtest",
            ],
            "tests": ["tests/testing/crashtest/crashtests.list",],
            "run_filename": "runreftestb2g.py",
            "testsdir": "reftest"
        },
        "jsreftest": {
            "options": [
                "--adbpath=%(adbpath)s",
                "--b2gpath=%(b2gpath)s",
                "--emulator=%(emulator)s",
                "--emulator-res=800x1000",
                "--logdir=%(logcat_dir)s",
                "--remote-webserver=%(remote_webserver)s",
                "--ignore-window-size",
                "--xre-path=%(xre_path)s",
                "--symbols-path=%(symbols_path)s",
                "--busybox=%(busybox)s",
                "--total-chunks=%(total_chunks)s",
                "--this-chunk=%(this_chunk)s",
                "--extra-profile-file=jsreftest/tests/user.js",
            ],
            "tests": ["jsreftest/tests/jstests.list",],
            "run_filename": "remotereftest.py",
            "testsdir": "reftest"
        },
        "mochitest": {
            "options": [
                "--adbpath=%(adbpath)s",
                "--b2gpath=%(b2gpath)s",
                "--emulator=%(emulator)s",
                "--logdir=%(logcat_dir)s",
                "--remote-webserver=%(remote_webserver)s",
                "--xre-path=%(xre_path)s",
                "--utility-path=%(utility_path)s",
                "--symbols-path=%(symbols_path)s",
                "--busybox=%(busybox)s",
                "--total-chunks=%(total_chunks)s",
                "--this-chunk=%(this_chunk)s",
                "--quiet",
                "--log-raw=%(raw_log_file)s",
                "--log-errorsummary=%(error_summary_file)s",
                "--certificate-path=%(certificate_path)s",
                "--screenshot-on-fail",
            ],
            "tests": ["%(test_path)s"],
            "run_filename": "runtestsb2g.py",
            "testsdir": "mochitest"
        },
        "mochitest-chrome": {
            "options": [
                "--adbpath=%(adbpath)s",
                "--b2gpath=%(b2gpath)s",
                "--emulator=%(emulator)s",
                "--logdir=%(logcat_dir)s",
                "--remote-webserver=%(remote_webserver)s",
                "--xre-path=%(xre_path)s",
                "--utility-path=%(utility_path)s",
                "--symbols-path=%(symbols_path)s",
                "--busybox=%(busybox)s",
                "--total-chunks=%(total_chunks)s",
                "--this-chunk=%(this_chunk)s",
                "--quiet",
                "--chrome",
                "--log-raw=%(raw_log_file)s",
                "--log-errorsummary=%(error_summary_file)s",
                "--certificate-path=%(certificate_path)s",
                "--screenshot-on-fail",
            ],
            "tests": ["%(test_path)s"],
            "run_filename": "runtestsb2g.py",
            "testsdir": "mochitest"
        },
        "reftest": {
            "options": [
                "--adbpath=%(adbpath)s",
                "--b2gpath=%(b2gpath)s",
                "--emulator=%(emulator)s",
                "--emulator-res=800x1000",
                "--logdir=%(logcat_dir)s",
                "--remote-webserver=%(remote_webserver)s",
                "--ignore-window-size",
                "--xre-path=%(xre_path)s",
                "--symbols-path=%(symbols_path)s",
                "--busybox=%(busybox)s",
                "--total-chunks=%(total_chunks)s",
                "--this-chunk=%(this_chunk)s",
                "--enable-oop",
            ],
            "tests": ["tests/layout/reftests/reftest.list",],
            "run_filename": "runreftestsb2g.py",
            "testsdir": "reftest"
        },
        "xpcshell": {
            "options": [
                "--adbpath=%(adbpath)s",
                "--b2gpath=%(b2gpath)s",
                "--emulator=%(emulator)s",
                "--logdir=%(logcat_dir)s",
                "--manifest=tests/xpcshell.ini",
                "--use-device-libs",
                "--testing-modules-dir=%(modules_dir)s",
                "--symbols-path=%(symbols_path)s",
                "--busybox=%(busybox)s",
                "--total-chunks=%(total_chunks)s",
                "--this-chunk=%(this_chunk)s",
                "--log-raw=%(raw_log_file)s",
                "--log-errorsummary=%(error_summary_file)s",
            ],
            "run_filename": "runtestsb2g.py",
            "testsdir": "xpcshell"
        }
    },
    "vcs_output_timeout": 1760,
}



#!/usr/bin/env python
# ***** BEGIN LICENSE BLOCK *****
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
# ***** END LICENSE BLOCK *****
"""vcs_sync.py

hg<->git conversions.  Needs to support both the monolithic beagle/gecko.git
type conversions, as well as many-to-many (l10n, build repos, etc.)
"""

from copy import deepcopy
import mmap
import os
import pprint
import re
import sys
import time

try:
    import simplejson as json
    assert json
except ImportError:
    import json

sys.path.insert(1, os.path.dirname(os.path.dirname(sys.path[0])))

import mozharness
external_tools_path = os.path.join(
    os.path.abspath(os.path.dirname(os.path.dirname(mozharness.__file__))),
    'external_tools',
)

from mozharness.base.errors import HgErrorList, GitErrorList
from mozharness.base.log import INFO, ERROR, FATAL
from mozharness.base.python import VirtualenvMixin, virtualenv_config_options
from mozharness.base.transfer import TransferMixin
from mozharness.base.vcs.vcssync import VCSSyncScript
from mozharness.mozilla.tooltool import TooltoolMixin


# HgGitScript {{{1
class HgGitScript(VirtualenvMixin, TooltoolMixin, TransferMixin, VCSSyncScript):
    """ Beagle-oriented hg->git script (lots of mozilla-central hardcodes;
        assumption that we're going to be importing lots of branches).

        Beagle is a git repo of mozilla-central, with full cvs history,
        and a number of developer-oriented repositories and branches added.

        The partner-oriented gecko.git could also be incorporated into this
        script with some changes.
        """

    mapfile_binary_search = None
    all_repos = None
    successful_repos = []
    config_options = [
        [["--no-check-incoming", ], {
            "action": "store_false",
            "dest": "check_incoming",
            "default": True,
            "help": "Don't check for incoming changesets",
        }],
        [["--max-log-sample-size", ], {
            "action": "store",
            "dest": "email_max_log_sample_size",
            "type": "int",
            "default": 102400,
            "help": "Specify the maximum number of characters from a log file to be "
                    "embedded in the email body, per embedding (not total - note we "
                    "embed two separate log samples into the email - so maximum "
                    "email body size can end up a little over 2x this amount).",
         }],
    ]

    def __init__(self, require_config_file=True):
        super(HgGitScript, self).__init__(
            config_options=virtualenv_config_options + self.config_options,
            all_actions=[
                'clobber',
                'list-repos',
                'create-virtualenv',
                'update-stage-mirror',
                'update-work-mirror',
                'create-git-notes',
                'publish-to-mapper',
                'push',
                'combine-mapfiles',
                'upload',
                'notify',
            ],
            # These default actions are the update loop that we run after the
            # initial steps to create the work mirror with all the branches +
            # cvs history have been run.
            default_actions=[
                'list-repos',
                'create-virtualenv',
                'update-stage-mirror',
                'update-work-mirror',
                'push',
                'combine-mapfiles',
                'upload',
                'notify',
            ],
            require_config_file=require_config_file
        )
        self.remote_targets = None

    # Helper methods {{{1
    def query_abs_dirs(self):
        """ Define paths.
            """
        if self.abs_dirs:
            return self.abs_dirs
        abs_dirs = super(HgGitScript, self).query_abs_dirs()
        abs_dirs['abs_cvs_history_dir'] = os.path.join(
            abs_dirs['abs_work_dir'], 'mozilla-cvs-history')
        abs_dirs['abs_source_dir'] = os.path.join(
            abs_dirs['abs_work_dir'], 'stage_source')
        abs_dirs['abs_repo_sync_tools_dir'] = os.path.join(
            abs_dirs['abs_work_dir'], 'repo-sync-tools')
        abs_dirs['abs_git_rewrite_dir'] = os.path.join(
            abs_dirs['abs_work_dir'], 'mc-git-rewrite')
        abs_dirs['abs_target_dir'] = os.path.join(abs_dirs['abs_work_dir'],
                                                  'target')
        if 'conversion_dir' in self.config:
            abs_dirs['abs_conversion_dir'] = os.path.join(
                abs_dirs['abs_work_dir'], 'conversion',
                self.config['conversion_dir']
            )
        self.abs_dirs = abs_dirs
        return self.abs_dirs

    def init_git_repo(self, path, additional_args=None, deny_deletes=False):
        """ Create a git repo, with retries.

            We call this with additional_args=['--bare'] to save disk +
            make things cleaner.
            """
        git = self.query_exe("git", return_type="list")
        cmd = git + ['init']
        # generally for --bare
        if additional_args:
            cmd.extend(additional_args)
        cmd.append(path)
        status = self.retry(
            self.run_command,
            args=(cmd, ),
            error_level=FATAL,
            error_message="Can't set up %s!" % path
        )
        status = self.run_command(
            git + ['config', 'receive.denyNonFastForwards', 'true'],
            cwd=path
        )
        if deny_deletes:
            status = self.run_command(
                git + ['config', 'receive.denyDeletes', 'true'],
                cwd=path
            )
        return status

    def write_hggit_hgrc(self, dest):
        """ Update .hg/hgrc, if not already updated
            """
        hgrc = os.path.join(dest, '.hg', 'hgrc')
        contents = ''
        if os.path.exists(hgrc):
            contents = self.read_from_file(hgrc)
        if 'hggit=' not in contents:
            hgrc_update = """[extensions]
hggit=
[git]
intree=1
"""
            self.write_to_file(hgrc, hgrc_update, open_mode='a')

    def _process_locale(self, locale, type, config, l10n_remote_targets, name, l10n_repos):
        """ This contains the common processing that we do on both gecko_config
            and gaia_config for a given locale.
            """
        replace_dict = {'locale': locale}
        new_targets = deepcopy(config.get('targets', {}))
        for target in new_targets:
            dest = target['target_dest']
            if '%(locale)s' in dest:
                new_dest = dest % replace_dict
                target['target_dest'] = new_dest
                remote_target = l10n_remote_targets.get(new_dest)
                if remote_target is None:  # generate target if not seen before
                    possible_remote_target = l10n_remote_targets.get(dest)
                    # target may be remote, or local - we can tell by seeing
                    # local targets will be none - remote targets will not,
                    # so we can use this to test if it is local or remote
                    if possible_remote_target is not None:
                        remote_target = deepcopy(possible_remote_target)
                        remote_repo = remote_target.get('repo')
                        if '%(locale)s' in remote_repo:
                            remote_target['repo'] = remote_repo % replace_dict
                        l10n_remote_targets[new_dest] = remote_target
        long_name = '%s_%s_%s' % (type, name, locale)
        repo_dict = {
            'repo': config['hg_url'] % replace_dict,
            'revision': 'default',
            'repo_name': long_name,
            'conversion_dir': long_name,
            'mapfile_name': '%s-mapfile' % long_name,
            'targets': new_targets,
            'vcs': 'hg',
            'branch_config': {
                'branches': {
                    'default': config['git_branch_name'],
                },
            },
            'tag_config': config.get('tag_config', {}),
            'mapper': config.get('mapper', {}),
            'generate_git_notes': config.get('generate_git_notes', {}),
        }
        l10n_repos.append(repo_dict)

    def _query_l10n_repos(self):
        """ Since I didn't want to have to build a huge static list of l10n
            repos, and since it would be nicest to read the list of locales
            from their SSoT files.
            """
        l10n_repos = []
        l10n_remote_targets = deepcopy(self.config['remote_targets'])
        dirs = self.query_abs_dirs()
        gecko_dict = deepcopy(self.config['l10n_config'].get('gecko_config', {}))
        for name, gecko_config in gecko_dict.items():
            file_name = self.download_file(gecko_config['locales_file_url'],
                                           parent_dir=dirs['abs_work_dir'])
            if not os.path.exists(file_name):
                self.error("Can't download locales from %s; skipping!" % gecko_config['locales_file_url'])
                continue
            contents = self.read_from_file(file_name)
            for locale in contents.splitlines():
                self._process_locale(locale, 'gecko', gecko_config, l10n_remote_targets, name, l10n_repos)

        gaia_dict = deepcopy(self.config['l10n_config'].get('gaia_config', {}))
        for name, gaia_config in gaia_dict.items():
            contents = self.retry(
                self.load_json_from_url,
                args=(gaia_config['locales_file_url'],)
            )
            if not contents:
                self.error("Can't download locales from %s; skipping!" % gaia_config['locales_file_url'])
                continue
            for locale in dict(contents).keys():
                self._process_locale(locale, 'gaia', gaia_config, l10n_remote_targets, name, l10n_repos)
        self.info("Built l10n_repos...")
        self.info(pprint.pformat(l10n_repos, indent=4))
        self.info("Remote targets...")
        self.info(pprint.pformat(l10n_remote_targets, indent=4))
        self.remote_targets = l10n_remote_targets
        return l10n_repos

    def _query_project_repos(self):
        """ Since I didn't want to have to build a huge static list of project
            branch repos.
            """
        project_repos = []
        for project in self.config.get("project_branches", []):
            repo_dict = {
                'repo': self.config['project_branch_repo_url'] % {'project': project},
                'revision': 'default',
                'repo_name': project,
                'targets': [{
                    'target_dest': 'github-project-branches',
                }],
                'vcs': 'hg',
                'branch_config': {
                    'branches': {
                        'default': project,
                    },
                },
                'tag_config': {},
            }
            project_repos.append(repo_dict)
        self.info("Built project_repos...")
        self.info(pprint.pformat(project_repos, indent=4))
        return project_repos

    def query_all_repos(self):
        """ Very simple method, but we need this concatenated list many times
            throughout the script.
            """
        if self.all_repos:
            return self.all_repos
        if self.config.get('conversion_type') == 'b2g-l10n':
            self.all_repos = self._query_l10n_repos()
        elif self.config.get('initial_repo'):
            self.all_repos = [self.config['initial_repo']] + list(self.config.get('conversion_repos', []))
        else:
            self.all_repos = list(self.config.get('conversion_repos', []))
        if self.config.get('conversion_type') == 'project-branches':
            self.all_repos += self._query_project_repos()
        return self.all_repos

    def query_all_non_failed_repos(self):
        """ Same as query_all_repos(self) but filters out repos that failed in an earlier
            action - so use this for downstream actions that require earlier actions did
            not fail for a given repo.
            """
        all_repos = self.query_all_repos()
        return [repo for repo in all_repos if repo.get('repo_name') not in self.failures]

    def _query_repo_previous_status(self, repo_name, repo_map=None):
        """ Return False if previous run was unsuccessful.
            Return None if no previous run information.
            """
        if repo_map is None:
            repo_map = self._read_repo_update_json()
        return repo_map.get('repos', {}).get(repo_name, {}).get('previous_push_successful')

    def _update_repo_previous_status(self, repo_name, successful_flag, repo_map=None, write_update=False):
        """ Set the repo_name to successful_flag (False for unsuccessful, True for successful)
            """
        if repo_map is None:
            repo_map = self._read_repo_update_json()
        repo_map.setdefault('repos', {}).setdefault(repo_name, {})['previous_push_successful'] = successful_flag
        if write_update:
            self._write_repo_update_json(repo_map)
        return repo_map

    def _update_stage_repo(self, repo_config, retry=True, clobber=False):
        """ Update a stage repo.
            See update_stage_mirror() for a description of the stage repos.
            """
        hg = self._query_hg_exe()
        dirs = self.query_abs_dirs()
        repo_name = repo_config['repo_name']
        source_dest = os.path.join(dirs['abs_source_dir'],
                                   repo_name)
        if clobber:
            self.rmtree(source_dest)
        if not os.path.exists(source_dest):
            if self.retry(
                self.run_command,
                args=(hg + ['clone', '--noupdate', repo_config['repo'],
                      source_dest], ),
                kwargs={
                    'output_timeout': 15 * 60,
                    'cwd': dirs['abs_work_dir'],
                    'error_list': HgErrorList,
                },
            ):
                if retry:
                    return self._update_stage_repo(
                        repo_config, retry=False, clobber=True)
                else:
                    # Don't leave a failed clone behind
                    self.rmtree(source_dest)
                    self._update_repo_previous_status(repo_name, successful_flag=False, write_update=True)
                    self.add_failure(
                        repo_name,
                        message="Can't clone %s!" % repo_config['repo'],
                        level=ERROR,
                    )
        elif self.config['check_incoming'] and repo_config.get("check_incoming", True):
            previous_status = self._query_repo_previous_status(repo_name)
            if previous_status is None:
                self.info("No previous status for %s; skipping incoming check!" % repo_name)
            elif previous_status is False:
                self.info("Previously unsuccessful status for %s; skipping incoming check!" % repo_name)
            else:
                # Run |hg incoming| and skip all subsequent actions if there
                # are no no changes.
                # If you want to bypass this behavior (e.g. to update branches/tags
                # on a repo without requiring a new commit), set
                # repo_config["incoming_check"] = False.
                cmd = hg + ['incoming', '-n', '-l', '1']
                status = self.retry(
                    self.run_command,
                    args=(cmd, ),
                    kwargs={
                        'output_timeout': 5 * 60,
                        'cwd': source_dest,
                        'error_list': HgErrorList,
                        'success_codes': [0, 1, 256],
                    },
                )
                if status in (1, 256):
                    self.info("No changes for %s; skipping." % repo_name)
                    # Overload self.failures to tell downstream actions to noop on
                    # this repo
                    self.failures.append(repo_name)
                    return
                elif status != 0:
                    self.add_failure(
                        repo_name,
                        message="Error getting changes for %s; skipping!" % repo_config['repo_name'],
                        level=ERROR,
                    )
                    self._update_repo_previous_status(repo_name, successful_flag=False, write_update=True)
                    return
        cmd = hg + ['pull']
        if self.retry(
            self.run_command,
            args=(cmd, ),
            kwargs={
                'output_timeout': 15 * 60,
                'cwd': source_dest,
                'error_list': HgErrorList,
            },
        ):
            if retry:
                return self._update_stage_repo(
                    repo_config, retry=False, clobber=True)
            else:
                self._update_repo_previous_status(repo_name, successful_flag=False, write_update=True)
                self.add_failure(
                    repo_name,
                    message="Can't pull %s!" % repo_config['repo'],
                    level=ERROR,
                )
        # commenting out hg verify since it takes ~5min per repo; hopefully
        # exit codes will save us
#        if self.run_command(hg + ["verify"], cwd=source_dest):
#            if retry:
#                return self._update_stage_repo(repo_config, retry=False, clobber=True)
#            else:
#                self.fatal("Can't verify %s!" % source_dest)

    def _do_push_repo(self, base_command, refs_list=None, kwargs=None):
        """ Helper method for _push_repo() since it has to be able to break
            out of the target_repo list loop, and the commands loop borks that.
            """
        commands = []
        if refs_list:
            while len(refs_list) > 10:
                commands.append(base_command + refs_list[0:10])
                refs_list = refs_list[10:]
            commands.append(base_command + refs_list)
        else:
            commands = [base_command]
        if kwargs is None:
            kwargs = {}
        for command in commands:
            # Do the push, with retry!
            if self.retry(
                self.run_command,
                args=(command, ),
                kwargs=kwargs,
            ):
                return -1

    def _push_repo(self, repo_config):
        """ Push a repo to a path ("test_push") or remote server.

            This was meant to be a cross-vcs method, but currently only
            covers git pushes.
            """
        dirs = self.query_abs_dirs()
        conversion_dir = self.query_abs_conversion_dir(repo_config)
        if not conversion_dir:
            self.fatal("No conversion_dir for %s!" % repo_config['repo_name'])
        source_dir = os.path.join(dirs['abs_source_dir'], repo_config['repo_name'])
        git = self.query_exe('git', return_type='list')
        hg = self._query_hg_exe()
        return_status = ''
        for target_config in repo_config['targets']:
            test_push = False
            remote_config = {}
            if target_config.get("test_push"):
                test_push = True
                force_push = target_config.get("force_push")
                target_name = os.path.join(
                    dirs['abs_target_dir'], target_config['target_dest'])
                target_vcs = target_config.get("vcs")
            else:
                target_name = target_config['target_dest']
                if self.remote_targets is None:
                    self.remote_targets = self.config.get('remote_targets', {})
                remote_config = self.remote_targets.get(target_name, target_config)
                force_push = remote_config.get("force_push", target_config.get("force_push"))
                target_vcs = remote_config.get("vcs", target_config.get("vcs"))
            if target_vcs == "git":
                base_command = git + ['push']
                env = {}
                if force_push:
                    base_command.append("-f")
                if test_push:
                    target_git_repo = target_name
                else:
                    target_git_repo = remote_config['repo']
                    # Allow for using a custom git ssh key.
                    env['GIT_SSH_KEY'] = remote_config['ssh_key']
                    env['GIT_SSH'] = os.path.join(external_tools_path, 'git-ssh-wrapper.sh')
                base_command.append(target_git_repo)
                # Allow for pushing a subset of repo branches to the target.
                # If we specify that subset, we can also specify different
                # names for those branches (e.g. b2g18 -> master for a
                # standalone b2g18 repo)
                # We query hg for these because the conversion dir will have
                # branches from multiple hg repos, and the regexes may match
                # too many things.
                refs_list = []
                if repo_config.get('generate_git_notes', False):
                    refs_list.append('+refs/notes/commits:refs/notes/commits')
                branch_map = self.query_branches(
                    target_config.get('branch_config', repo_config.get('branch_config', {})),
                    source_dir,
                )
                # If the target_config has a branch_config, the key is the
                # local git branch and the value is the target git branch.
                if target_config.get("branch_config"):
                    for (branch, target_branch) in branch_map.items():
                        refs_list += ['+refs/heads/%s:refs/heads/%s' % (branch, target_branch)]
                # Otherwise the key is the hg branch and the value is the git
                # branch; use the git branch for both local and target git
                # branch names.
                else:
                    for (hg_branch, git_branch) in branch_map.items():
                        refs_list += ['+refs/heads/%s:refs/heads/%s' % (git_branch, git_branch)]
                # Allow for pushing a subset of tags to the target, via name or
                # regex.  Again, query hg for this list because the conversion
                # dir will contain tags from multiple hg repos, and the regexes
                # may match too many things.
                tag_config = target_config.get('tag_config', repo_config.get('tag_config', {}))
                if tag_config.get('tags'):
                    for (tag, target_tag) in tag_config['tags'].items():
                        refs_list += ['+refs/tags/%s:refs/tags/%s' % (tag, target_tag)]
                if tag_config.get('tag_regexes'):
                    regex_list = []
                    for regex in tag_config['tag_regexes']:
                        regex_list.append(re.compile(regex))
                    tag_list = self.get_output_from_command(
                        hg + ['tags'],
                        cwd=source_dir,
                    )
                    if tag_list is not None:
                        for tag_line in tag_list.splitlines():
                            if not tag_line:
                                continue
                            tag_parts = tag_line.split()
                            if not tag_parts:
                                self.error("Bogus tag_line? %s" % str(tag_line))
                                continue
                            tag_name = tag_parts[0]
                            for regex in regex_list:
                                if tag_name != 'tip' and regex.search(tag_name) is not None:
                                    refs_list += ['+refs/tags/%s:refs/tags/%s' % (tag_name, tag_name)]
                                    continue
                error_msg = "%s: Can't push %s to %s!\n" % (repo_config['repo_name'], conversion_dir, target_git_repo)
                if self._do_push_repo(
                    base_command,
                    refs_list=refs_list,
                    kwargs={
                        'output_timeout': target_config.get("output_timeout", 30 * 60),
                        'cwd': os.path.join(conversion_dir, '.git'),
                        'error_list': GitErrorList,
                        'partial_env': env,
                    }
                ):
                    if target_config.get("test_push"):
                        error_msg += "This was a test push that failed; not proceeding any further with %s!\n" % repo_config['repo_name']
                    self.error(error_msg)
                    return_status += error_msg
                    if target_config.get("test_push"):
                        break
            else:
                # TODO write hg
                error_msg = "%s: Don't know how to deal with vcs %s!\n" % (
                    target_config['target_dest'], target_vcs)
                self.error(error_msg)
                return_status += error_msg
        return return_status

    def _query_mapped_revision(self, revision=None, mapfile=None):
        """ Use the virtualenv mapper module to search a mapfile for a
            revision.
            """
        if not callable(self.mapfile_binary_search):
            site_packages_path = self.query_python_site_packages_path()
            sys.path.append(os.path.join(site_packages_path, 'mapper'))
            try:
                from bsearch import mapfile_binary_search
                global log
                log = self.log_obj
                self.mapfile_binary_search = mapfile_binary_search
            except ImportError, e:
                self.fatal("Can't import mapfile_binary_search! %s\nDid you create-virtualenv?" % str(e))
        # I wish mapper did this for me, but ...
        fd = open(mapfile, 'rb')
        m = mmap.mmap(fd.fileno(), 0, mmap.MAP_PRIVATE, mmap.PROT_READ)
        return self.mapfile_binary_search(m, revision)

    def _post_fatal(self, message=None, exit_code=None):
        """ After we call fatal(), run this method before exiting.
            """
        if 'notify' in self.actions:
            self.notify(message=message, fatal=True)
        self.copy_logs_to_upload_dir()

    def _read_repo_update_json(self):
        """ repo_update.json is a file we create with information about each
            repo we're converting: git/hg branch names, git/hg revisions,
            pull datetime/timestamp, and push datetime/timestamp.

            Since we want to be able to incrementally update portions of this
            file as we pull/push each branch, we need to be able to read the
            json into memory, so we can update the dict and re-write the json
            to disk.
            """
        repo_map = {}
        dirs = self.query_abs_dirs()
        path = os.path.join(dirs['abs_upload_dir'], 'repo_update.json')
        if os.path.exists(path):
            fh = open(path, 'r')
            repo_map = json.load(fh)
            fh.close()
        return repo_map

    def query_abs_conversion_dir(self, repo_config):
        dirs = self.query_abs_dirs()
        if repo_config.get('conversion_dir'):
            dest = os.path.join(dirs['abs_work_dir'], 'conversion',
                                repo_config['conversion_dir'])
        else:
            dest = dirs.get('abs_conversion_dir')
        return dest

    def _write_repo_update_json(self, repo_map):
        """ The write portion of _read_repo_update_json().
            """
        dirs = self.query_abs_dirs()
        contents = json.dumps(repo_map, sort_keys=True, indent=4)
        self.write_to_file(
            os.path.join(dirs['abs_upload_dir'], 'repo_update.json'),
            contents,
            create_parent_dir=True
        )

    def _query_hg_exe(self):
        """Returns the hg executable command as a list
        """
        # If "hg" is set in "exes" section of config use that.
        # If not, get path from self.query_virtualenv_path() method
        # (respects --work-dir and --venv-path and --virtualenv-path).
        exe_command = self.query_exe('hg', return_type="list", default=[os.path.join(self.query_virtualenv_path(), "bin", "hg")])

        # possible additional command line options can be specified in "hg_options" of self.config
        hg_options = self.config.get("hg_options", ())
        exe_command.extend(hg_options)
        return exe_command

    def query_branches(self, branch_config, repo_path, vcs='hg'):
        """ Given a branch_config of branches and branch_regexes, return
            a dict of existing branch names to target branch names.
            """
        branch_map = {}
        if "branches" in branch_config:
            branch_map = deepcopy(branch_config['branches'])
        if "branch_regexes" in branch_config:
            regex_list = list(branch_config['branch_regexes'])
            full_branch_list = []
            if vcs == 'hg':
                hg = self._query_hg_exe()
                # This assumes we always want closed branches as well.
                # If not we may need more options.
                output = self.get_output_from_command(
                    hg + ['branches', '-a'],
                    cwd=repo_path
                )
                if output:
                    for line in output.splitlines():
                        full_branch_list.append(line.split()[0])
            elif vcs == 'git':
                git = self.query_exe("git", return_type="list")
                output = self.get_output_from_command(
                    git + ['branch', '-l'],
                    cwd=repo_path
                )
                if output:
                    for line in output.splitlines():
                        full_branch_list.append(line.replace('*', '').split()[0])
            for regex in regex_list:
                for branch in full_branch_list:
                    m = re.search(regex, branch)
                    if m:
                        # Don't overwrite branch_map[branch] if it exists
                        branch_map.setdefault(branch, branch)
        return branch_map

    def _combine_mapfiles(self, mapfiles, combined_mapfile, cwd=None):
        """ Adapted from repo-sync-tools/combine_mapfiles

            Consolidate multiple conversion processes' mapfiles into a
            single mapfile.
            """
        self.info("Determining whether we need to combine mapfiles...")
        if cwd is None:
            cwd = self.query_abs_dirs()['abs_upload_dir']
        existing_mapfiles = []
        for f in mapfiles:
            f_path = os.path.join(cwd, f)
            if os.path.exists(f_path):
                existing_mapfiles.append(f)
            else:
                self.warning("%s doesn't exist!" % f_path)
        combined_mapfile_path = os.path.join(cwd, combined_mapfile)
        if os.path.exists(combined_mapfile_path):
            combined_timestamp = os.path.getmtime(combined_mapfile_path)
            for f in existing_mapfiles:
                f_path = os.path.join(cwd, f)
                if os.path.getmtime(f_path) > combined_timestamp:
                    # Yes, we want to combine mapfiles
                    break
            else:
                self.info("No new mapfiles to combine.")
                return
            self.move(combined_mapfile_path, "%s.old" % combined_mapfile_path)
        output = self.get_output_from_command(
            ['sort', '--unique', '-t', ' ',
             '--key=2'] + existing_mapfiles,
            silent=True, halt_on_failure=True,
            cwd=cwd,
        )
        self.write_to_file(combined_mapfile_path, output, verbose=False,
                           error_level=FATAL)
        self.run_command(['ln', '-sf', combined_mapfile,
                          '%s-latest' % combined_mapfile],
                         cwd=cwd)

    # Actions {{{1

    def list_repos(self):
        repos = self.query_all_repos()
        self.info(pprint.pformat(repos))

    def create_test_targets(self):
        """ This action creates local directories to do test pushes to.
            """
        dirs = self.query_abs_dirs()
        for repo_config in self.query_all_non_failed_repos():
            for target_config in repo_config['targets']:
                if not target_config.get('test_push'):
                    continue
                target_dest = os.path.join(dirs['abs_target_dir'], target_config['target_dest'])
                if not os.path.exists(target_dest):
                    self.info("Creating local target repo %s." % target_dest)
                    if target_config.get("vcs", "git") == "git":
                        self.init_git_repo(target_dest, additional_args=['--bare', '--shared=all'],
                                           deny_deletes=True)
                    else:
                        self.fatal("Don't know how to deal with vcs %s!" % target_config['vcs'])
                else:
                    self.debug("%s exists; skipping." % target_dest)

    def update_stage_mirror(self):
        """ The stage mirror is a buffer clean clone of repositories.
            The logic behind this is that we get occasional corruption from
            |hg pull|.  It's much less time-consuming to detect this in
            a clean clone, and reclone, than to detect this in a working
            conversion directory, and try to repair or reclone+reconvert.

            We pull the stage mirror into the work mirror, where the conversion
            is done.
            """
        for repo_config in self.query_all_non_failed_repos():
            self._update_stage_repo(repo_config)

    def pull_out_new_sha_lookups(self, old_file, new_file):
        """ This method will return an iterator which iterates through lines in file
            new_file that do not exist in old_file. If old_file can't be read, all
            lines in new_file are returned. It does not cause any problems if lines
            exist in old_file that do not exist in new_file. Results are sorted by
            the second field (text after first space in line).

            This is somewhat equivalent to:
               ( [ ! -f "${old_file}" ] && cat "${new_file}" || diff "${old_file}" "${new_file}" | sed -n 's/> //p' ) | sort -k2"""
        with self.opened(old_file) as (old, err):
            if err:
                self.info('Map file %s not found - probably first time this has run.' % old_file)
                old_set = frozenset()
            else:
                old_set = frozenset(old)
        with self.opened(new_file, 'rt') as (new, err):
            if err:
                self.error('Could not read contents of map file %s:\n%s' % (new_file, err.message))
                new_set = frozenset()
            else:
                new_set = frozenset(new)
        for line in sorted(new_set.difference(old_set), key=lambda line: line.partition(' ')[2]):
            yield line

    def update_work_mirror(self):
        """ Pull the latest changes into the work mirror, update the repo_map
            json, and run |hg gexport| to convert those latest changes into
            the git conversion repo.
            """
        hg = self._query_hg_exe()
        git = self.query_exe("git", return_type="list")
        dirs = self.query_abs_dirs()
        repo_map = self._read_repo_update_json()
        timestamp = int(time.time())
        datetime = time.strftime('%Y-%m-%d %H:%M %Z')
        repo_map['last_pull_timestamp'] = timestamp
        repo_map['last_pull_datetime'] = datetime
        for repo_config in self.query_all_non_failed_repos():
            repo_name = repo_config['repo_name']
            source = os.path.join(dirs['abs_source_dir'], repo_name)
            dest = self.query_abs_conversion_dir(repo_config)
            if not dest:
                self.fatal("No conversion_dir for %s!" % repo_name)
            if not os.path.exists(dest):
                self.mkdir_p(os.path.dirname(dest))
                self.run_command(hg + ['clone', '--noupdate', source, dest],
                                 error_list=HgErrorList,
                                 halt_on_failure=False)
                if os.path.exists(dest):
                    self.write_hggit_hgrc(dest)
                    self.init_git_repo('%s/.git' % dest, additional_args=['--bare'])
                    self.run_command(
                    git + ['--git-dir', '%s/.git' % dest, 'config', 'gc.auto', '0'],
                    )
                else:
                    self.add_failure(
                        repo_name,
                        message="Failed to clone %s!" % source,
                        level=ERROR,
                    )
                    continue
            # Build branch map.
            branch_map = self.query_branches(
                repo_config.get('branch_config', {}),
                source,
            )
            for (branch, target_branch) in branch_map.items():
                output = self.get_output_from_command(
                    hg + ['id', '-r', branch],
                    cwd=source
                )
                if output:
                    rev = output.split(' ')[0]
                else:
                    self.add_failure(
                        repo_name,
                        message="Branch %s doesn't exist in %s (%s cloned into staging directory %s)!" % (branch, repo_name, repo_config.get('repo'), source),
                        level=ERROR,
                    )
                    continue
                timestamp = int(time.time())
                datetime = time.strftime('%Y-%m-%d %H:%M %Z')
                if self.run_command(hg + ['pull', '-r', rev, source], cwd=dest,
                                    error_list=HgErrorList):
                    # We shouldn't have an issue pulling!
                    self.add_failure(
                        repo_name,
                        message="Unable to pull %s from stage_source; clobbering and skipping!" % repo_name,
                        level=ERROR,
                    )
                    self._update_repo_previous_status(repo_name, successful_flag=False, write_update=True)
                    # don't leave a dirty checkout behind, and skip remaining branches
                    self.rmtree(source)
                    break
                self.run_command(
                    hg + ['bookmark', '-f', '-r', rev, target_branch],
                    cwd=dest, error_list=HgErrorList,
                )
                # This might get a little large.
                repo_map.setdefault('repos', {}).setdefault(repo_name, {}).setdefault('branches', {})[branch] = {
                    'hg_branch': branch,
                    'hg_revision': rev,
                    'git_branch': target_branch,
                    'pull_timestamp': timestamp,
                    'pull_datetime': datetime,
                }
            if self.query_failure(repo_name):
                # We hit an error in the for loop above
                continue
            self.retry(
                self.run_command,
                args=(hg + ['-v', 'gexport'], ),
                kwargs={
                    'output_timeout': 15 * 60,
                    'cwd': dest,
                    'error_list': HgErrorList,
                },
                error_level=FATAL,
            )
            generated_mapfile = os.path.join(dest, '.hg', 'git-mapfile')
            self.copy_to_upload_dir(
                generated_mapfile,
                dest=repo_config.get('mapfile_name', self.config.get('mapfile_name', "gecko-mapfile")),
                log_level=INFO
            )
            for (branch, target_branch) in branch_map.items():
                git_revision = self._query_mapped_revision(
                    revision=rev, mapfile=generated_mapfile)
                repo_map['repos'][repo_name]['branches'][branch]['git_revision'] = git_revision
        self._write_repo_update_json(repo_map)


    def create_git_notes(self):
        git = self.query_exe("git", return_type="list")
        for repo_config in self.query_all_non_failed_repos():
            repo = repo_config['repo']
            if repo_config.get('generate_git_notes', False):
                dest = self.query_abs_conversion_dir(repo_config)
                # 'git-mapfile' is created by hggit plugin, containing all the mappings
                complete_mapfile = os.path.join(dest, '.hg', 'git-mapfile')
                # 'added-to-git-notes' is the set of mappings known to be recorded in the git notes
                # of the project (typically 'git-mapfile' from previous run)
                added_to_git_notes = os.path.join(dest, '.hg', 'added-to-git-notes')
                # 'delta-git-notes' is the set of new mappings found on this iteration, that
                # now need to be added to the git notes of the project (the diff between the
                # previous two files described)
                delta_git_notes = os.path.join(dest, '.hg', 'delta-git-notes')
                git_dir = os.path.join(dest, '.git')
                self.rmtree(delta_git_notes)
                git_notes_adding_successful = True
                with self.opened(delta_git_notes, open_mode='w') as (delta_out, err):
                    if err:
                        git_notes_adding_successful = False
                        self.warn("Could not write list of unprocessed git note mappings to file %s - not critical" % delta_git_notes)
                    else:
                        for sha_lookup in self.pull_out_new_sha_lookups(added_to_git_notes, complete_mapfile):
                            print >>delta_out, sha_lookup,
                            (git_sha, hg_sha) = sha_lookup.split()
                            # only add git note if not already there - note
                            # devs may have added their own notes, so don't
                            # replace any existing notes, just add to them
                            output = self.get_output_from_command(
                                git + ['notes', 'show', git_sha],
                                cwd=git_dir,
                                ignore_errors=True
                            )
                            git_note_text='Upstream source: %s/rev/%s' % (repo, hg_sha)
                            git_notes_add_return_code = 1
                            if not output or output.find(git_note_text) < 0:
                                git_notes_add_return_code = self.run_command(
                                    git + ['notes', 'append', '-m', git_note_text, git_sha],
                                    cwd=git_dir
                                )
                            # if note was successfully added, or it was already there, we can
                            # mark it as added, by putting it in the delta file...
                            if git_notes_add_return_code == 0 or output.find(git_note_text) >= 0:
                                print >>delta_out, sha_lookup,
                            else:
                                self.error("Was not able to append required git note for git commit %s ('%s')" % (git_sha, git_note_text))
                                git_notes_adding_successful = False
                if git_notes_adding_successful:
                    self.copyfile(complete_mapfile, added_to_git_notes)
            else:
                self.info("Not creating git notes for repo %s (generate_git_notes not set to True)" % repo)

    def publish_to_mapper(self):
        """ This method will attempt to create git notes for any new git<->hg mappings
            found in the generated_mapfile file and also push new mappings to mapper service."""
        for repo_config in self.query_all_non_failed_repos():
            dest = self.query_abs_conversion_dir(repo_config)
            # 'git-mapfile' is created by hggit plugin, containing all the mappings
            complete_mapfile = os.path.join(dest, '.hg', 'git-mapfile')
            # 'published-to-mapper' is all the mappings that are known to be published
            # to mapper, for this project (typically the 'git-mapfile' from the previous
            # run)
            published_to_mapper = os.path.join(dest, '.hg', 'published-to-mapper')
            # 'delta-for-mapper' is the set of mappings that need to be published to
            # mapper on this iteration, i.e. the diff between the previous two files
            # described
            delta_for_mapper = os.path.join(dest, '.hg', 'delta-for-mapper')
            self.rmtree(delta_for_mapper)
            # we only replace published_to_mapper if we successfully updated
            # pushed to mapper
            mapper_config = repo_config.get('mapper', {})
            if mapper_config:
                site_packages_path = self.query_python_site_packages_path()
                if site_packages_path not in sys.path:
                    sys.path.append(site_packages_path)
                try:
                    import requests
                except ImportError as e:
                    self.error("Can't import requests: %s\nDid you create-virtualenv?" % str(e))
                mapper_url = mapper_config['url']
                mapper_project = mapper_config['project']
                insert_url = "%s/%s/insert/ignoredups" % (mapper_url, mapper_project)
                headers = {
                    'Content-Type': 'text/plain',
                    'Authentication': 'Bearer %s' % os.environ["RELENGAPI_INSERT_HGGIT_MAPPINGS_AUTH_TOKEN"]
                }
                all_new_mappings = []
                all_new_mappings.extend(self.pull_out_new_sha_lookups(published_to_mapper, complete_mapfile))
                self.write_to_file(delta_for_mapper, "".join(all_new_mappings))
                # due to timeouts on load balancer, we only push 200 lines at a time
                # this means that we should get http response back within 30 seconds
                # including the time it takes to insert the mappings in the database
                publish_successful = True
                for i in range(0, len(all_new_mappings), 200):
                    r = requests.post(insert_url, data="".join(all_new_mappings[i:i+200]), headers=headers)
                    if (r.status_code != 200):
                        self.error("Could not publish mapfile ('%s') line range [%s, %s] to mapper (%s) - received http %s code" % (delta_for_mapper, i, i+200, insert_url, r.status_code))
                        publish_successful = False
                        # we won't break out, since we may be able to publish other mappings
                        # and duplicates are allowed, so we will push the whole lot again next
                        # time anyway
                    else:
                        self.info("Published mapfile ('%s') line range [%s, %s] to mapper (%s)" % (delta_for_mapper, i, i+200, insert_url))
                if publish_successful:
                    # if we get this far, we know we could successfully post to mapper, so now
                    # we can copy the mapfile over "previously generated" version
                    # so that we don't push to mapper for these commits again
                    self.copyfile(complete_mapfile, published_to_mapper)
            else:
                self.copyfile(complete_mapfile, published_to_mapper)

    def combine_mapfiles(self):
        """ This method is for any job (l10n, project-branches) that needs to combine
            mapfiles.
            """
        if not self.config.get("combined_mapfile"):
            self.info("No combined_mapfile set in config; skipping!")
            return
        dirs = self.query_abs_dirs()
        mapfiles = []
        if self.config.get('conversion_type') == 'b2g-l10n':
            for repo_config in self.query_all_non_failed_repos():
                if repo_config.get("mapfile_name"):
                    mapfiles.append(repo_config['mapfile_name'])
        else:
            mapfiles.append(self.config.get('mapfile_name', 'gecko-mapfile'))
        if self.config.get('external_mapfile_urls'):
            for url in self.config['external_mapfile_urls']:
                file_name = self.download_file(
                    url,
                    parent_dir=dirs['abs_upload_dir'],
                    error_level=FATAL,
                )
                mapfiles.append(file_name)
        if not mapfiles:
            self.info("No mapfiles to combine; skipping!")
            return
        self._combine_mapfiles(mapfiles, self.config['combined_mapfile'])

    def push(self):
        """ Push to all targets.  test_targets are local directory test repos;
            the rest are remote.  Updates the repo_map json.
            """
        self.create_test_targets()
        repo_map = self._read_repo_update_json()
        failure_msg = ""
        timestamp = int(time.time())
        datetime = time.strftime('%Y-%m-%d %H:%M %Z')
        repo_map['last_push_timestamp'] = timestamp
        repo_map['last_push_datetime'] = datetime
        for repo_config in self.query_all_non_failed_repos():
            timestamp = int(time.time())
            datetime = time.strftime('%Y-%m-%d %H:%M %Z')
            status = self._push_repo(repo_config)
            repo_name = repo_config['repo_name']
            if not status:  # good
                if repo_name not in self.successful_repos:
                    self.successful_repos.append(repo_name)
                repo_map.setdefault('repos', {}).setdefault(repo_name, {})['push_timestamp'] = timestamp
                repo_map['repos'][repo_name]['push_datetime'] = datetime
                previous_status = self._query_repo_previous_status(repo_name, repo_map=repo_map)
                if previous_status is None:
                    self.add_summary("Possibly the first successful push of %s." % repo_name)
                elif previous_status is False:
                    self.add_summary("Previously unsuccessful push of %s is now successful!" % repo_name)
                self._update_repo_previous_status(repo_name, successful_flag=True, repo_map=repo_map, write_update=True)
            else:
                self.add_failure(
                    repo_name,
                    message="Unable to push %s." % repo_name,
                    level=ERROR,
                )
                failure_msg += status + "\n"
                self._update_repo_previous_status(repo_name, successful_flag=False, repo_map=repo_map, write_update=True)
        if not failure_msg:
            repo_map['last_successful_push_timestamp'] = repo_map['last_push_timestamp']
            repo_map['last_successful_push_datetime'] = repo_map['last_push_datetime']
        self._write_repo_update_json(repo_map)
        if failure_msg:
            self.fatal("Unable to push these repos:\n%s" % failure_msg)

    def preflight_upload(self):
        if not self.config.get("copy_logs_post_run", True):
            self.copy_logs_to_upload_dir()

    def upload(self):
        """ Upload the upload_dir according to the upload_config.
            """
        failure_msg = ''
        dirs = self.query_abs_dirs()
        for upload_config in self.config.get('upload_config', []):
            if self.retry(
                self.rsync_upload_directory,
                args=(
                    dirs['abs_upload_dir'],
                ),
                kwargs=upload_config,
            ):
                failure_msg += '%s:%s' % (upload_config['remote_host'],
                                          upload_config['remote_path'])
        if failure_msg:
            self.fatal("Unable to upload to this location:\n%s" % failure_msg)


# __main__ {{{1
if __name__ == '__main__':
    conversion = HgGitScript()
    conversion.run()
